# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#

from dataclasses import dataclass
from typing import Optional

from rfmt.blocks import LineBlock as LB
from rfmt.blocks import TextBlock as TB
from rfmt.blocks import StackBlock as SB
from rfmt.blocks import ChoiceBlock as CB
from rfmt.blocks import IndentBlock as IB
from rfmt.blocks import WrapBlock as WB

from .utils import with_commas

from .node import SQLNodeList
from .node import SQLNode
from .query import SQLQuery
from .types import SQLType
from .expr import SQLExpr
from .expr import SQLBaseExpr


@dataclass(frozen=True)
class SQLBiOp(SQLExpr):
    sql_op: str
    left: SQLExpr
    right: SQLExpr

    def sqlf(self, compact):
        right_expr_compact = self.right.sqlf(True)
        left_expr_compact = self.left.sqlf(True)

        compact_sql = LB([
            left_expr_compact, TB(' '), TB(self.sql_op),
            TB(' '),
            right_expr_compact
        ])
        if compact:
            return compact_sql

        right_expr = self.right.sqlf(False)
        left_expr = self.left.sqlf(False)

        return CB([
            compact_sql, SB([
                LB([left_expr, TB(' '), TB(self.sql_op)]),
                right_expr
            ])
        ])

    @staticmethod
    def parse(lex):
        return SQLBiOp.parse_or_expression(lex)

    @staticmethod
    def parse_or_expression(lex):
        expr = SQLBiOp.parse_and_expression(lex)

        while lex.consume('OR'):
            expr = SQLBiOp('OR', expr, SQLBiOp.parse_and_expression(lex))

        return expr

    @staticmethod
    def parse_and_expression(lex):
        expr = SQLBiOp.parse_not_expression(lex)

        while lex.consume('AND'):
            expr = SQLBiOp('AND', expr, SQLBiOp.parse_not_expression(lex))

        return expr

    @staticmethod
    def parse_not_expression(lex):
        # Add while loop here.
        if lex.consume('NOT'):
            return SQLUniOp('NOT', SQLBiOp.parse_compare_expression(lex))

        return SQLBiOp.parse_compare_expression(lex)

    @staticmethod
    def consume_basic_cmp_operator(lex):
        return lex.consume_any(['!=', '<>', '<=>', '<=',
                                '>=', '=', '<', '>'])

    @staticmethod
    def consume_is_operator(lex):
        if not lex.consume('IS'):
            return None

        is_op = 'IS '
        if lex.consume('NOT'):
            is_op += 'NOT '

        typ = lex.consume_any(['TRUE', 'FALSE', 'NULL'])
        if typ:
            return is_op+typ

        if lex.consume('DISTINCT'):
            lex.expect('FROM')
            return is_op + ' DISTINCT FROM'

        lex.error('Expected TRUE, FALSE, NULL, or DISTINCT FROM')
        return None

    @staticmethod
    def parse_compare_expression(lex):
        expr = SQLBiOp.parse_bit_or_expression(lex)

        while True:
            sql_op = SQLBiOp.consume_basic_cmp_operator(lex)
            if sql_op:
                expr = SQLBiOp(sql_op, expr,
                               SQLBiOp.parse_bit_or_expression(lex))
                continue

            is_op = SQLBiOp.consume_is_operator(lex)
            if is_op:
                expr = SQLUniOp(is_op, expr)
                continue

            inverted = ''
            if lex.consume('NOT'):
                inverted = 'NOT '

            if lex.consume('LIKE'):
                expr = SQLLike(inverted + 'LIKE', expr,
                               SQLBiOp.parse_bit_or_expression(lex))
            elif lex.consume('BETWEEN'):
                expr_l = SQLBiOp.parse_bit_or_expression(lex)
                lex.expect('AND')
                expr_r = SQLBiOp.parse_bit_or_expression(lex)
                expr = SQLBetween(inverted + 'BETWEEN',
                                  expr, expr_l, expr_r)
            elif lex.consume('IN'):
                lex.expect('(')
                in_query = SQLQuery.consume(lex)
                if in_query:
                    lex.expect(')')
                    return SQLINSQL(inverted + 'IN', expr, in_query)
                vals = []
                while True:
                    vals.append(SQLBiOp.parse_bit_or_expression(lex))
                    if not lex.consume(','):
                        break
                lex.expect(')')
                return SQLIN(inverted + 'IN', expr, SQLNodeList(vals))
            else:
                break

        return expr

    @staticmethod
    def parse_bit_or_expression(lex) -> SQLExpr:
        expr = SQLBiOp.parse_bit_xor_expression(lex)

        while lex.consume('|'):
            expr = SQLBiOp('|', expr, SQLBiOp.parse_bit_xor_expression(lex))

        return expr

    @staticmethod
    def parse_bit_xor_expression(lex) -> SQLExpr:
        expr = SQLBiOp.parse_bit_and_expression(lex)

        while lex.consume('^'):
            expr = SQLBiOp('^', expr, SQLBiOp.parse_bit_and_expression(lex))

        return expr

    @staticmethod
    def parse_bit_and_expression(lex):
        expr = SQLBiOp.parse_bit_shift_expression(lex)

        while lex.consume('&'):
            expr = SQLBiOp('&', expr,
                           SQLBiOp.parse_bit_shift_expression(lex))

        return expr

    @staticmethod
    def parse_bit_shift_expression(lex):
        expr = SQLBiOp.parse_add_sub_expression(lex)

        while True:
            if lex.consume('<<'):
                expr = SQLBiOp('<<', expr,
                               SQLBiOp.parse_add_sub_expression(lex))
            elif lex.consume('>>'):
                expr = SQLBiOp('>>', expr,
                               SQLBiOp.parse_add_sub_expression(lex))
            else:
                break

        return expr

    @staticmethod
    def parse_add_sub_expression(lex):
        expr = SQLBiOp.parse_mul_div_expression(lex)

        while True:
            if lex.consume('+'):
                expr = SQLBiOp('+', expr,
                               SQLBiOp.parse_mul_div_expression(lex))
            elif lex.consume('-'):
                expr = SQLBiOp('-', expr,
                               SQLBiOp.parse_mul_div_expression(lex))
            else:
                break

        return expr

    @staticmethod
    def parse_mul_div_expression(lex):
        expr = SQLBiOp.parse_str_expression(lex)

        while True:
            if lex.consume('*'):
                expr = SQLBiOp('*', expr,
                               SQLBiOp.parse_str_expression(lex))
            elif lex.consume('/'):
                expr = SQLBiOp('/', expr,
                               SQLBiOp.parse_str_expression(lex))
            else:
                break

        return expr

    @staticmethod
    def parse_str_expression(lex):
        expr = SQLUniOp.parse(lex)

        while True:
            if lex.consume('||'):
                expr = SQLBiOp('||', expr,
                               SQLUniOp.parse(lex))
            else:
                break

        return expr


# Pass in priority for expanding on the brackets.
@dataclass(frozen=True)
class SQLUniOp(SQLExpr):
    sql_op: str
    arg: SQLExpr

    def sqlf(self, compact):
        if self.sql_op.startswith('IS'):
            return LB(
                [self.arg.sqlf(compact), TB(' '), TB(self.sql_op)])

        if self.sql_op.isalpha():
            return LB(
                [TB(self.sql_op), TB(' '), self.arg.sqlf(compact)])

        return LB([TB(self.sql_op), self.arg.sqlf(compact)])

    @staticmethod
    def parse(lex):
        if lex.consume('-'):
            expr = SQLUniOp('-', SQLCase.parse(lex))
        elif lex.consume('~'):
            expr = SQLUniOp('~', SQLCase.parse(lex))
        else:
            expr = SQLCase.parse(lex)

        return expr


@dataclass(frozen=True)
class SQLBetween(SQLExpr):
    sql_op: str
    bexpr: SQLExpr
    left: SQLExpr
    right: SQLExpr

    def sqlf(self, compact):
        return LB([
            self.bexpr.sqlf(compact), TB(' '), TB(self.sql_op),
            TB(' '), self.left.sqlf(compact), TB(' '),
            TB('AND'), TB(' '), self.right.sqlf(compact)
        ])


@dataclass(frozen=True)
class SQLLike(SQLExpr):
    sql_op: str
    lexpr: SQLExpr
    match: SQLExpr

    def sqlf(self, compact):
        return LB([
            self.lexpr.sqlf(True), TB(' ' + self.sql_op + ' '),
            self.match.sqlf(True)
        ])


@dataclass(frozen=True)
class SQLINSQL(SQLExpr):
    sql_op: str
    expr: SQLExpr
    sql: SQLNode

    def sqlf(self, compact):
        sql_op = TB(' ' + self.sql_op + ' (')
        compact_sql = LB([self.expr.sqlf(True), sql_op] +
                         [self.sql.sqlf(True), TB(')')])
        if compact:
            return compact_sql

        return CB([
            compact_sql,
            SB([
                LB([self.expr.sqlf(False), sql_op]),
                IB(self.sql.sqlf(True)),
                TB(')')
            ]),
        ])


@dataclass(frozen=True)
class SQLIN(SQLExpr):
    sql_op: str
    iexpr: SQLExpr
    args: SQLNodeList[SQLExpr]

    def sqlf(self, compact):
        in_vals = []
        for arg in self.args[:-1]:
            in_vals.append(LB([arg.sqlf(True), TB(',')]))
        in_vals.append(self.args[-1].sqlf(True))
        sql_op = TB(' ' + self.sql_op + ' (')
        compact_sql = LB([self.iexpr.sqlf(True), sql_op] + in_vals +
                         [TB(')')])
        if compact:
            return compact_sql

        # TODO(scannell): Bug with separator - incorrect usage
        return CB([
            compact_sql,
            SB([
                LB([self.iexpr.sqlf(False), sql_op]),
                IB(WB(in_vals, sep=' ')),
                TB(')')
            ]),
        ])


@dataclass(frozen=True)
class SQLCase(SQLExpr):
    base_expr: Optional[SQLExpr]
    else_expr: Optional[SQLExpr]
    args: SQLNodeList

    def sqlf(self, compact):
        big_block = []
        small_block = []
        for if_then in zip(self.args[0::2], self.args[1::2]):
            if_expr = if_then[0].sqlf(False)
            if_expr_compact = if_then[0].sqlf(True)
            then_expr = if_then[1].sqlf(False)
            then_expr_compact = if_then[1].sqlf(True)

            # Most compact form - one line
            small_block.append(
                LB([
                    TB(' WHEN '), if_expr_compact, TB(' THEN '),
                    then_expr_compact
                ]))

            # Choice
            big_block.append(
                CB([
                    SB([
                        TB('WHEN'),
                        IB(if_expr),
                        TB('THEN'),
                        IB(then_expr)
                    ]),
                    SB([
                        LB([TB('WHEN '), if_expr_compact, TB(' THEN')]),
                        IB(then_expr)
                    ])
                ])
            )
        if self.else_expr:
            else_expr_compact = LB([TB(' ELSE '), self.else_expr.sqlf(True)])
            small_block.append(else_expr_compact)

            else_expr_norm = self.else_expr.sqlf(False)
            big_block.append(
                CB([
                    SB([TB('ELSE '), IB(else_expr_norm)]),
                    else_expr_compact
                ])
            )

        case_block = TB('CASE ')
        case_block_compact = TB('CASE')
        if self.base_expr:
            case_block_compact = LB([
                TB('CASE '), self.base_expr.sqlf(True)
            ])
            case_block = LB([
                TB('CASE '), self.base_expr.sqlf(False), TB(' ')
            ])

        compact_sql = LB([case_block_compact] + small_block +
                         [TB(' END')])
        if compact:
            return compact_sql

        return CB([
            compact_sql,
            SB([case_block, IB(SB(big_block)), TB('END')])
        ])

    @staticmethod
    def parse(lex):
        if not lex.consume('CASE'):
            return SQLColonCast.parse(lex)

        base_expr: Optional[SQLExpr] = None
        if not lex.peek('WHEN'):
            base_expr = SQLExpr.parse(lex)

        else_expr: Optional[SQLExpr] = None
        args = []
        while True:
            if lex.consume('WHEN'):
                args.append(SQLExpr.parse(lex))
                lex.expect('THEN')
                args.append(SQLExpr.parse(lex))
                continue
            if lex.consume('ELSE'):
                else_expr = SQLExpr.parse(lex)
                lex.expect('END')
                break
            if lex.consume('END'):
                break
            lex.error('Expected WHEN, ELSE, or END')
        return SQLCase(base_expr, else_expr, SQLNodeList(args))


@dataclass(frozen=True)
class SQLColonCast(SQLExpr):
    arg: SQLExpr
    typ: SQLType

    def sqlf(self, compact):
        return LB([
            self.arg.sqlf(True),
            TB('::'),
            self.typ.sqlf(True),
        ])

    @staticmethod
    def parse(lex) -> SQLExpr:
        expr = (SQLBrackets.consume(lex) or
                SQLBaseExpr.parse(lex))

        if lex.consume('::'):
            typ = SQLType.parse(lex)
            expr = SQLColonCast(expr, typ)

        return expr


@dataclass(frozen=True)
class SQLStruct(SQLExpr):
    exprs: SQLNodeList[SQLExpr]

    def sqlf(self, compact):
        compact_sql = LB([TB('(')] +
                         with_commas(True, self.exprs, tail=')'))
        if compact:
            return compact_sql
        return CB([
            compact_sql,
            SB([
                TB('('),
                WB(with_commas(False, self.exprs, tail=')')),
            ])
        ])


@dataclass(frozen=True)
class SQLBrackets(SQLExpr):
    query: SQLNode

    def sqlf(self, compact):
        compact_sql = LB([
            TB('('),
            self.query.sqlf(True),
            TB(')')
        ])

        if compact:
            return compact_sql

        return CB([
            compact_sql,
            SB([
                TB('('),
                IB(self.query.sqlf(compact)),
                TB(')')
            ])
        ])

    @staticmethod
    def consume(lex) -> Optional[SQLExpr]:

        if not lex.consume('('):
            return None

        # If peek of SELECT or WITH then a sub-select it is
        if lex.peek('WITH') or lex.peek('SELECT'):
            query = SQLQuery.parse(lex)
            lex.expect(')')
            return SQLBrackets(query)

        exprs = []
        while True:
            sub_expr = SQLExpr.parse(lex)
            exprs.append(sub_expr)
            if not lex.consume(','):
                break

        lex.expect(')')

        if len(exprs) == 1:
            return SQLBrackets(exprs[0])

        return SQLStruct(SQLNodeList(exprs))



import unittest
from sql_refactor import Refactor

class TestRefactor(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super(TestRefactor, self).__init__(*args, **kwargs)

        knowledge = {
                        'table_a':
                        {
                            'new_table' : 'new_table_a',
                            'column_knowledge':
                            {
                                'column_1' : 'new_column_1',
                                'column_2' : 'new_column_2',
                                'column_3' : 'new_column_3',
                            },
                            'preserved' : False
                        },
                        'table_b':
                        {
                            'new_table' : 'new_table_b',
                            'column_knowledge':
                            {
                                'column_1' : 'new_column_1b',
                                'column_2b' : 'new_column_2b',
                            },
                            'preserved' : False
                        },
                        'table_c':
                        {
                            'new_table' : 'new_table_c',
                            'column_knowledge':
                            {
                                'column_1c' : 'new_column_1c',
                            },
                            'preserved' : False
                        },
                        'table_d':
                        {
                            'new_table' : None,
                            'column_knowledge': 
                            {
                                'column_1' : None,
                                'column_2' : None,
                            },
                            'preserved' : True
                        }
                    }
        self.command = Refactor(knowledge)

    def _assert_equal_sql(self, sql, reference):
        self.command.refactor(sql)
        refactored = self.command.result()
        refactored = refactored.replace('`', '')

        self.command.refactor(reference, parse_only=True)
        reference = self.command.result()
        reference = reference.replace('`', '')

        return self.assertEqual(refactored, reference)

    def test_simple_select(self):
        sql = """
        SELECT column_1, column_2
        FROM table_a
        """

        reference = """
        SELECT new_column_1 AS column_1, new_column_2 AS column_2
        FROM new_table_a
        """

        self._assert_equal_sql(sql, reference)

    def test_select_where(self):
        sql = """
        SELECT column_1, column_2
        FROM table_a
        WHERE column_1 = 1
        """

        reference = """
        SELECT new_column_1 AS column_1, new_column_2 AS column_2
        FROM new_table_a
        WHERE new_column_1 = 1
        """

        self._assert_equal_sql(sql, reference)

    def test_column_not_found(self):
        sql = """
        SELECT column_1, column_0
        FROM table_a
        """

        reference = """
        SELECT 
            new_column_1 AS column_1, 
            -- [WARNING] Could not find column in knowledge: column_0
            column_0
        FROM new_table_a
        """

        self._assert_equal_sql(sql, reference)

    def test_table_not_found(self):
        sql = """
        SELECT column_1
        FROM table_x
        """

        reference = """
        -- [WARNING] Could not find table in knowledge: table_x
        SELECT
            column_1
        FROM table_x
        """

        self._assert_equal_sql(sql, reference)

    def test_field_function(self):
        sql = """
        SELECT 
        CAST(column_1 AS STRING) AS new_column_1, column_2
        FROM table_a
        WHERE 
        CAST(column_1 AS STRING) = "column_name"
        """

        reference = """
        SELECT
        CAST(new_column_1 AS STRING) AS new_column_1,
        new_column_2 AS column_2
        FROM
        `new_table_a`
        WHERE
        CAST(new_column_1 AS STRING) = 'column_name'
        """

        self._assert_equal_sql(sql, reference)

    def test_join(self):
        # LEFT JOIN ON
        sql = """
        SELECT 
        CAST(a.column_1 AS STRING) AS new_column_1,
        column_2,
        b.column_1 AS column_1b,
        column_2b,
        FROM table_a AS a
        LEFT JOIN table_b AS b ON a.column_1 = b.column_1
        WHERE 
        CAST(b.column_1 AS STRING) = "column_name"
        """

        reference = """
        SELECT 
        CAST(a.new_column_1 AS STRING) AS new_column_1,
        a.new_column_2 AS column_2,
        b.new_column_1b AS column_1b,
        b.new_column_2b AS column_2b,
        FROM new_table_a AS a
        LEFT JOIN new_table_b AS b ON a.new_column_1 = b.new_column_1b
        WHERE 
        CAST(b.new_column_1b AS STRING) = 'column_name'
        """

        self._assert_equal_sql(sql, reference)

        # CROSS JOIN
        sql = """
        SELECT 
        column_1, column_1c
        FROM table_a, table_c
        WHERE 
        CAST(column_1 AS STRING) = "column_name"
        """

        reference = """
        SELECT 
        new_column_1 AS column_1,
        new_column_1c AS column_1c
        FROM new_table_a, new_table_c
        WHERE 
        CAST(new_column_1 AS STRING) = "column_name"
        """

        self._assert_equal_sql(sql, reference)

        # LEFT JOIN USING
        sql = """
        SELECT 
        CAST(a.column_1 AS STRING) AS new_column_1,
        column_2,
        b.column_1 AS column_1b,
        column_2b,
        FROM table_a AS a
        LEFT JOIN table_b AS b USING(column_1)
        WHERE 
        CAST(b.column_1 AS STRING) = "column_name"
        """

        reference = """
        SELECT 
        CAST(a.new_column_1 AS STRING) AS new_column_1,
        a.new_column_2 AS column_2,
        b.new_column_1b AS column_1b,
        b.new_column_2b AS column_2b,
        FROM new_table_a AS a
        LEFT JOIN new_table_b AS b ON a.new_column_1 = b.new_column_1b
        WHERE 
        CAST(b.new_column_1b AS STRING) = 'column_name'
        """

        self._assert_equal_sql(sql, reference)

        # LEFT JOIN USING, WITH
        sql = """
        WITH tb AS(
            SELECT column_1, column_2b
            FROM table_b
        )
        SELECT 
        CAST(a.column_1 AS STRING) AS new_column_1,
        column_2,
        tb.column_1 AS column_1b,
        column_2b,
        FROM table_a AS a
        LEFT JOIN tb USING(column_1)
        WHERE 
        CAST(tb.column_1 AS STRING) = "column_name"
        """

        reference = """
        WITH tb AS(
            SELECT new_column_1b AS column_1, new_column_2b AS column_2b
            FROM new_table_b
        )
        SELECT 
        CAST(a.new_column_1 AS STRING) AS new_column_1,
        a.new_column_2 AS column_2,
        tb.column_1 AS column_1b,
        tb.column_2b,
        FROM new_table_a AS a
        LEFT JOIN tb AS tb ON a.new_column_1 = tb.column_1
        WHERE 
        CAST(tb.column_1 AS STRING) = 'column_name'
        """

        self._assert_equal_sql(sql, reference)

    def test_preserved(self):
        sql = """
        SELECT column_1, column_2
        FROM table_d
        """

        reference = """
        SELECT column_1, column_2
        FROM table_d
        """

        self._assert_equal_sql(sql, reference)

    def test_with(self):
        sql = """
        WITH cte_a AS(
            SELECT column_1, column_2
            FROM table_a
        )
        SELECT column_1, column_2 FROM cte_a
        """

        reference = """
        WITH cte_a AS(
            SELECT new_column_1 AS column_1, new_column_2 AS column_2
            FROM new_table_a
        )
        SELECT column_1, column_2 FROM cte_a
        """

        self._assert_equal_sql(sql, reference)

        sql = """
        WITH cte_a AS(
            WITH cte_a AS(
                SELECT column_1 AS c1, column_2
                FROM table_a
            )
            SELECT c1 AS column_1, column_2
            FROM cte_a
        )
        SELECT column_1, column_2 FROM cte_a
        """

        reference = """
        WITH cte_a AS(
            WITH cte_a AS(
                SELECT new_column_1 AS c1, new_column_2 AS column_2
                FROM new_table_a
            )
            SELECT c1 AS column_1, column_2
            FROM cte_a
        )
        SELECT column_1, column_2 FROM cte_a
        """

        self._assert_equal_sql(sql, reference)

    def test_temp_table(self):
        sql = """
        CREATE TEMP TABLE cte_a AS
        SELECT column_1, column_2 FROM table_a;
        
        SELECT column_1, column_2
        FROM cte_a
        """

        reference = """
        CREATE TEMP TABLE cte_a AS
        SELECT new_column_1 AS column_1,
        new_column_2 AS column_2 
        FROM new_table_a;
        
        SELECT column_1, column_2
        FROM cte_a
        """

        self._assert_equal_sql(sql, reference)

        sql = """
        CREATE TEMP TABLE cte_a AS
        SELECT column_1, column_2 FROM table_a;
        
        CREATE TEMP TABLE cte_b AS
        SELECT column_1, column_2b AS column_2 from table_b;

        SELECT a.column_1, b.column_2
        FROM cte_a AS a
        LEFT JOIN cte_b AS b
        USING(column_1)
        """

        reference = """
        CREATE TEMP TABLE cte_a AS
        SELECT new_column_1 AS column_1, new_column_2 AS column_2 FROM new_table_a;
        
        CREATE TEMP TABLE cte_b AS
        SELECT new_column_1b AS column_1, new_column_2b AS column_2 from new_table_b;

        SELECT a.column_1, b.column_2
        FROM cte_a AS a
        LEFT JOIN cte_b AS b
        USING(column_1)
        """

        self._assert_equal_sql(sql, reference)


    def test_wildcard(self):
        sql = """
        SELECT * FROM table_a
        """

        reference = """
        SELECT new_column_1 AS column_1,
        new_column_2 AS column_2,
        new_column_3 AS column_3
        FROM new_table_a
        """

        self._assert_equal_sql(sql, reference)

        sql = """
        SELECT * EXCEPT(column_2) FROM table_a
        """

        reference = """
        SELECT new_column_1 AS column_1,
        new_column_3 AS column_3
        FROM new_table_a
        """

        self._assert_equal_sql(sql, reference)

        sql = """
        SELECT a.* EXCEPT(column_2) , b.column_2b
        FROM table_a As a
        LEFT JOIN table_b AS b
        ON a.column_1 = b.column_1
        """

        reference = """
        SELECT  
        a.new_column_1 AS column_1,
        a.new_column_3 AS column_3,
        b.new_column_2b AS column_2b
        FROM new_table_a As a
        LEFT JOIN new_table_b AS b
        ON a.new_column_1 = b.new_column_1b
        """

        self._assert_equal_sql(sql, reference)

        sql = """
        SELECT b.column_2b, a.* EXCEPT(column_2)
        FROM table_a As a
        LEFT JOIN table_b AS b
        ON a.column_1 = b.column_1
        """

        reference = """
        SELECT  
        b.new_column_2b AS column_2b,
        a.new_column_1 AS column_1,
        a.new_column_3 AS column_3
        FROM new_table_a As a
        LEFT JOIN new_table_b AS b
        ON a.new_column_1 = b.new_column_1b
        """

        self._assert_equal_sql(sql, reference)

    def test_nested_select(self):
        sql = """
        SELECT column_1
        FROM(
            SELECT * FROM table_a
        )
        """

        reference = """
        SELECT column_1
        FROM(
            SELECT new_column_1 AS column_1,
            new_column_2 AS column_2,
            new_column_3 AS column_3,
            FROM new_table_a
        )
        """

        self._assert_equal_sql(sql, reference)

        sql = """
        SELECT a.column_1, b.column_2b
        FROM table_b AS b
        LEFT JOIN (
            SELECT * FROM table_a
        ) a
        USING(column_1)
        """

        reference = """
        SELECT a.column_1, b.new_column_2b AS column_2b
        FROM new_table_b AS b
        LEFT JOIN (
            SELECT new_column_1 AS column_1,
            new_column_2 AS column_2,
            new_column_3 AS column_3,
            FROM new_table_a
        ) AS a
        ON b.new_column_1b = a.column_1
        """

        self._assert_equal_sql(sql, reference)

    def test_union(self):
        sql = """
        SELECT column_1 FROM table_a
        UNION ALL
        SELECT column_1 FROM table_b
        """

        reference = """
        SELECT new_column_1 AS column_1 FROM `new_table_a`
        UNION ALL
        SELECT new_column_1b AS column_1 FROM `new_table_b`
        """

        self._assert_equal_sql(sql, reference)

        sql = """
        SELECT column_1
        FROM(
            SELECT column_1 FROM table_a
            UNION ALL
            SELECT column_1 FROM table_b
        )
        """

        reference = """
        SELECT column_1
        FROM(
            SELECT new_column_1 AS column_1 FROM `new_table_a`
            UNION ALL
            SELECT new_column_1b AS column_1 FROM `new_table_b`
        )
        """

        self._assert_equal_sql(sql, reference)

    def test_ordered_query(self):
        sql = """
        WITH ta AS (
            SELECT column_1, column_2
            FROM table_a
            order by column_1
        )
        SELECT *
        FROM ta
        """

        reference = """
        WITH ta AS (
            SELECT new_column_1 AS column_1, new_column_2 AS column_2
            FROM new_table_a
            order by column_1
        )
        SELECT column_1, column_2
        FROM ta
        """

        self._assert_equal_sql(sql, reference)


# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#

import unittest
import mock

from .const import SQLConstant

from . import parse


def mock_type_parser(lex):
    lex.expect('TYPE')
    return SQLConstant('TYPE')


def mock_parse_expression(lex):
    return SQLConstant(lex.parse_identifier())


class TestParser(unittest.TestCase):

    def assert_sql(self, val, fsql):
        expr = parse(val)
        self.assertEqual(str(expr), fsql)

        nsql = expr.as_sql(False)
        nexpr = parse(nsql)
        self.assertEqual(str(expr), str(nexpr))

    def test_basic_sql(self):
        self.assert_sql('/* x */ select E1 as x, E2, E3 as i',
                        'SELECT E1 AS x,E2,E3 AS i')

    def test_select(self):
        self.assert_sql('/* x */ select EXPR as x, EXPR, EXPR as i',
                        'SELECT EXPR AS x,EXPR,EXPR AS i')
        self.assert_sql('select EXPR as fieldA where EXPR group by fieldA',
                        'SELECT EXPR AS fieldA WHERE EXPR GROUP BY fieldA')
        self.assert_sql('/* x */ select EXPR as x, EXPR, EXPR as i',
                        'SELECT EXPR AS x,EXPR,EXPR AS i')
        self.assert_sql('SELECT EXPR UNION ALL SELECT EXPR',
                        'SELECT EXPR UNION ALL SELECT EXPR')

    def test_withselect(self):
        self.assert_sql('with x as (select E1) select E2 as fieldA '
                        'where WEXPR group by fieldA',
                        'WITH x AS(SELECT E1)SELECT E2 AS fieldA '
                        'WHERE WEXPR GROUP BY fieldA')

    # TODO(scannell): Generate SQL for user defined functions as well!
    @mock.patch('sql_parser.types.SQLType.parse',
                mock.MagicMock(side_effect=mock_type_parser))
    def test_function(self):
        self.assert_sql('CREATE TEMPORARY FUNCTION TO_STRING(x TYPE) AS'
                        ' FExpr; SELECT F;',
                        'CREATE TEMPORARY FUNCTION TO_STRING(x TYPE)'
                        'AS FExpr;SELECT F')
        self.assert_sql('CREATE TEMPORARY FUNCTION stuff(x TYPE) '
                        'RETURNS TYPE '
                        'LANGUAGE js AS """ javascript stuff """;'
                        'SELECT F;',
                        'CREATE TEMPORARY FUNCTION stuff(x TYPE)'
                        'RETURNS TYPE LANGUAGE '
                        'js AS \' javascript stuff \';SELECT F')

    def test_order_by(self):
        self.assert_sql('SELECT x FROM t ORDER BY tx,tb ASC',
                        'SELECT x FROM t ORDER BY tx,tb ASC')
        self.assert_sql('SELECT x FROM t ORDER BY tx DESC,tb ASC',
                        'SELECT x FROM t ORDER BY tx DESC,tb ASC')

    def test_limit_offset(self):
        self.assert_sql('SELECT x FROM t LIMIT 10', 'SELECT x FROM t LIMIT 10')
        self.assert_sql('SELECT x FROM t LIMIT 10 OFFSET 10',
                        'SELECT x FROM t LIMIT 10 OFFSET 10')

    def test_query_table(self):
        self.assert_sql('SELECT E1 FROM(SELECT stuff FROM t)f',
                        'SELECT E1 FROM(SELECT stuff FROM t)AS f')
        self.assert_sql('SELECT E1, FROM(SELECT stuff FROM t)f',
                        'SELECT E1 FROM(SELECT stuff FROM t)AS f')

    def test_sub_query(self):
        self.assert_sql('(SELECT F1 FROM stuff)ORDER BY F2',
                        '(SELECT F1 FROM stuff)ORDER BY F2')
        self.assert_sql('(SELECT F1 FROM stuff ORDER BY f)ORDER BY F2',
                        '(SELECT F1 FROM stuff ORDER BY f)ORDER BY F2')
        self.assert_sql('select F1 FROM stuff ORDER BY F2',
                        'SELECT F1 FROM stuff ORDER BY F2')

    def test_having(self):
        self.assert_sql('SELECT Fx FROM t GROUP BY Tx HAVING E1',
                        'SELECT Fx FROM t GROUP BY Tx HAVING E1')

    def test_select_type(self):
        self.assert_sql('SELECT DISTINCT f FROM t', 'SELECT DISTINCT f FROM t')
        self.assert_sql('SELECT ALL f FROM t', 'SELECT ALL f FROM t')

    def test_sets(self):
        self.assert_sql('SELECT x FROM t UNION ALL SELECT x FROM t',
                        'SELECT x FROM t UNION ALL SELECT x FROM t')
        self.assert_sql('SELECT x FROM t UNION DISTINCT SELECT x FROM t',
                        'SELECT x FROM t UNION DISTINCT SELECT x FROM t')
        self.assert_sql('SELECT x FROM t INTERSECT DISTINCT SELECT x FROM t',
                        'SELECT x FROM t INTERSECT DISTINCT SELECT x FROM t')
        self.assert_sql('SELECT x FROM t EXCEPT DISTINCT SELECT x FROM t',
                        'SELECT x FROM t EXCEPT DISTINCT SELECT x FROM t')

    # TODO(scannell)
    # Sub-query as table
    def test_select_expr(self):
        self.assert_sql('SELECT X FROM Y WHERE F IN (SELECT A FROM B)',
                        'SELECT X FROM Y WHERE F IN(SELECT A FROM B)')
        self.assert_sql('SELECT (SELECT 1)',
                        'SELECT(SELECT 1)')

    def test_join(self):
        self.assert_sql('SELECT X FROM Table Y JOIN Table X ON Expr1',
                        'SELECT X FROM Table AS Y INNER JOIN Table AS X '
                        'ON Expr1')
        self.assert_sql('SELECT X FROM Table Y LEFT JOIN Table X',
                        'SELECT X FROM Table AS Y LEFT JOIN Table AS X')
        self.assert_sql('SELECT X FROM Table Y INNER JOIN Table X',
                        'SELECT X FROM Table AS Y INNER JOIN Table AS X')
        self.assert_sql('SELECT X FROM Table Y CROSS JOIN Table X',
                        'SELECT X FROM Table AS Y CROSS JOIN Table AS X')
        self.assert_sql('SELECT X FROM Table Y, Table X',
                        'SELECT X FROM Table AS Y CROSS JOIN Table AS X')
        self.assert_sql('SELECT X FROM Table Y LEFT OUTER JOIN Table X',
                        'SELECT X FROM Table AS Y LEFT OUTER JOIN Table AS X')
        self.assert_sql('SELECT X FROM Table Y RIGHT OUTER JOIN Table X',
                        'SELECT X FROM Table AS Y RIGHT OUTER JOIN Table AS X')
        self.assert_sql('SELECT X FROM Table Y RIGHT JOIN Table X',
                        'SELECT X FROM Table AS Y RIGHT JOIN Table AS X')
        self.assert_sql('SELECT X FROM Table Y FULL OUTER JOIN Table X',
                        'SELECT X FROM Table AS Y FULL OUTER JOIN Table AS X')
        self.assert_sql('SELECT X FROM Table Y FULL JOIN Table X',
                        'SELECT X FROM Table AS Y FULL JOIN Table AS X')
        self.assert_sql('SELECT X FROM Table Y LEFT JOIN Table X USING(a)',
                        'SELECT X FROM Table AS Y LEFT JOIN Table AS X ON Y.a=X.a')
        self.assert_sql('SELECT X FROM Table Y, UNNEST(A) Z',
                        'SELECT X FROM Table AS Y CROSS JOIN UNNEST(A)AS Z')

    def test_truncate(self):
        self.assert_sql('TRUNCATE  TABLE /* b */ tt;',
                        'TRUNCATE TABLE tt;')
        self.assert_sql('truncate table td_address',
                        'TRUNCATE TABLE td_address;')

    def test_multiple(self):
        self.assert_sql('TRUNCATE  TABLE /* b */ tt;  SELECT  1;',
                        'TRUNCATE TABLE tt;SELECT 1')

    def test_insert(self):
        self.assert_sql('INSERT INTO TABLE (a,b,c) SELECT 1,2;',
                        'INSERT INTO TABLE(a,b,c)SELECT 1,2;')
        self.assert_sql('INSERT INTO TABLE (SELECT 1,2);',
                        'INSERT INTO TABLE(SELECT 1,2);')

    def test_generate_stats(self):
        self.assert_sql('GENERATE   STATistics on blah;',
                        'GENERATE STATISTICS ON blah;')
        self.assert_sql('GROOM TABLE x RECORDS ALL;',
                        'GROOM TABLE x RECORDS ALL;')

    def test_update(self):
        self.assert_sql('UPDATE x SET f1=1, f2=a+b FROM y WHERE TRUE',
                        'UPDATE x SET f1=1,f2=a+b FROM y WHERE TRUE')

    def test_trans(self):
        self.assert_sql('BEGIN;', 'BEGIN;')

    def test_sql_expr(self):
        self.assert_sql('select ((select 1) + 2)',
                        'SELECT((SELECT 1)+2)')

    def test_sql_delete(self):
        self.assert_sql('DELETE from y where true',
                        'DELETE FROM y WHERE TRUE;')

    def test_sql_create(self):
        self.assert_sql('CREATE OR replace TABLE y (i INT64, b BOOL) AS SELECT 1',
                        'CREATE OR REPLACE TABLE y(i INT64,b BOOL)AS SELECT 1')



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#

"""SQL Base Parser - parse basic constructs from a StringReader.

"""

#####################################################################
#
# General tools for for examining, consuming, and expecting elements
#

from typing import List


class ParsingError(Exception):
    pass


# 'IF' isn't reserved... treat it as a normal internal function.
RESERVED_WORDS = (['ALL', 'AND', 'ANY', 'ARRAY', 'AS', 'ASC',
                   'ASSERT_ROWS_MODIFIED', 'AT', 'BETWEEN', 'BY', 'CASE',
                   'CAST', 'COLLATE', 'CONTAINS', 'CREATE', 'CROSS',
                   'CUBE', 'CURRENT', 'DEFAULT', 'DEFINE', 'DESC',
                   'DISTINCT', 'ELSE', 'END', 'ENUM', 'ESCAPE', 'EXCEPT',
                   'EXCLUDE', 'EXISTS', 'EXTRACT', 'FALSE', 'FETCH',
                   'FOLLOWING', 'FOR', 'FROM', 'FULL', 'GROUP', 'GROUPING',
                   'GROUPS', 'HASH', 'HAVING', 'IGNORE', 'IN', 'INNER',
                   'INTERSECT', 'INTERVAL', 'INTO', 'IS', 'JOIN',
                   'LATERAL', 'LEFT', 'LIKE', 'LIMIT', 'LOOKUP', 'MERGE',
                   'NATURAL', 'NEW', 'NO', 'NOT', 'NULL', 'NULLS', 'OF',
                   'ON', 'OR', 'ORDER', 'OUTER', 'OVER', 'PARTITION',
                   'PRECEDING', 'PROTO', 'RANGE', 'RECURSIVE', 'RESPECT',
                   'RIGHT', 'ROLLUP', 'ROWS', 'SELECT', 'SET', 'SOME',
                   'STRUCT', 'TABLESAMPLE', 'THEN', 'TO', 'TREAT', 'TRUE',
                   'UNBOUNDED', 'UNION', 'UNNEST', 'USING', 'WHEN', 'MINUS',
                   'WHERE', 'WINDOW', 'WITH', 'WITHIN', 'QUALIFY', 'RECORDS'])


class SQLLexer:

    def __init__(self, sql_str):
        self._str = sql_str
        self._pos = 0
        self._comments = []

        self.consume_all_space()

    CONTEXT = 25

    def error(self, msg):

        # Find beginning and end of snippet.
        pos = self._pos
        linec = self._str[0:pos].count('\n') + 1
        line_start = 0
        if linec > 1:
            line_start = self._str[0:pos + 1].rindex('\n') + 1

        snippet_beg = int(max(line_start, pos - SQLLexer.CONTEXT))
        try:
            snippet_end = int(
                min(self._str[line_start:].index('\n') + line_start + 1,
                    pos + SQLLexer.CONTEXT))
        except ValueError:
            snippet_end = int(min(len(self._str), pos + SQLLexer.CONTEXT))

        # Raise exception
        raise ParsingError('\n'.join([
            f'{msg} at [{linec}:{pos - line_start + 1}]',
            f'Snippet: {self._str[snippet_beg:snippet_end]}',
            f'         {"-" * (pos - snippet_beg)}^'
        ]))

    def get_comments(self) -> List[str]:
        rcomments = self._comments
        self._comments = []
        return rcomments

    def peek(self, elem):
        """Peek for an element.

        Args:
          elem: Element. If a string, an identifier is searched for. Otherwise
              an operator.

        Returns:
          Returns True/False if the element is available to consume.
        """
        if isinstance(elem, list):
            rpos = self._pos
            for el in elem:
                if not self.peek(el):
                    self._pos = rpos
                    return False
                self._pos += len(el)
                self.consume_all_space()
            self._pos = rpos
            return True

        # If elem does not start with an alpha character,
        # just check if it matches the next available text
        if not elem[0].isalpha():
            return self._str[self._pos:self._pos + len(elem)] == elem

        # If it *doesn't* match the upper case of next available next,
        # then it doesn't match.
        if self._str[self._pos:self._pos + len(elem)].upper() != elem:
            return False

        # If this is the end of the string, it's a match
        if self._pos + len(elem) >= len(self._str):
            return True

        # Check the next character -- if it's not an alpha character,
        # and not a digit or underscore then we have a match.
        ch = self._str[self._pos + len(elem)]
        if not ch.isalpha() and not ch.isdigit() and ch != '_':
            return True

        return False

    def consume_any(self, elems):
        """Consume any of a list of elements.

        Args:
            elems: List of elements
                See peek() for what is an element
                Tested by consume(elem)

        Returns:
            Returns elem from elems that was consumed or None
        """
        for e in elems:
            if self.consume(e):
                return e
        return None

    def consume(self, elem):
        """Consume for an element or list of elements.

        Args:
          elem(s): Element or list of elements.
              See peek() for what is an element.
              If a list, then all elements must be consumed to be satisfied.

        Returns:
          Returns True/False if the element(s) is/are available to consume.
        """
        if isinstance(elem, list):
            rpos = self._pos
            for e in elem:
                if not self.consume(e):
                    self._pos = rpos
                    return None
            return elem

        if self.peek(elem):
            self._pos += len(elem)
            self.consume_all_space()
            return elem

        return None

    def expect(self, elem):
        return (self.consume(elem) or
                self.error('Expected "' + elem + '"'))

    def expect_end(self):
        if self._pos < len(self._str):
            self.error('Expected end')

    def consume_all_space(self):
        """Consume all space and comments.

        Returns:
          This moves the position forward to the next non-space non-comment.
        """
        while self._pos < len(self._str):

            if self._str[self._pos].isspace():
                self._pos += 1
                continue

            # It's a SQL comment
            if self._str[self._pos:self._pos + 2] == '--':
                start = self._pos + 2
                try:
                    move = self._str[self._pos:].index('\n')
                except ValueError:
                    move = len(self._str) - self._pos
                self._comments.append(self._str[start:start + move - 2].strip())
                self._pos += move
                continue

            # It's a hash comment
            if self._str[self._pos] == '#':
                try:
                    move = self._str[self._pos:].index('\n')
                except ValueError:
                    move = len(self._str) - self._pos
                self._pos += move
                continue

            # It's a /* */ comment
            if self._str[self._pos:self._pos + 2] == '/*':
                start = self._pos + 2
                while self._pos < len(self._str):
                    if self._str[self._pos:self._pos + 2] == '*/':
                        self._comments.append(
                            self._str[start:self._pos].strip())
                        self._pos += 2
                        break
                    self._pos += 1
                continue

            # Something else - stop now!
            break

    def consume_identifier(self):
        """Consume identifier.

        Returns:
          This returns an identifier (that's not reserved) or None if
          one isn't available.
        """
        if self._pos >= len(self._str):
            return None

        is_escaped = None
        ch = self._str[self._pos]
        if ch in ('`', '"'):
            is_escaped = ch

        if not (is_escaped or
                ch.isalpha() or
                ch == ':' or
                ch == '_'):
            return None

        pos = self._pos

        if is_escaped:
            pos += 1

        v = []

        if ch == ':':
            v.append(ch)
            pos += 1

        while pos < len(self._str):
            ch = self._str[pos]
            if is_escaped:
                if ch == is_escaped:
                    break
            else:
                if not (ch.isalpha() or ch == '_' or ch.isdigit()):
                    break
            v.append(ch)
            pos += 1

        v = ''.join(v)
        if v.upper() in RESERVED_WORDS:
            if is_escaped:
                v = ''.join((is_escaped, v, is_escaped))
            else:
                return None

        if is_escaped:
            pos += 1

        self._pos = pos

        self.consume_all_space()

        return v

    def consume_number(self):
        """Consume number.

        Returns:
          This returns a number (float or integer) or None
          if one isn't available.
        """
        if self._pos == len(self._str):
            return None

        if not self._str[self._pos].isdigit():
            return None

        # Do the parsing itself.
        pos = self._pos
        is_frac = False
        divisor = 1
        num = 0
        while pos < len(self._str):
            ch = self._str[pos]
            if ch.isdigit():
                num *= 10
                num += int(ch)
                if is_frac:
                    divisor *= 10
            elif not is_frac and ch == '.':
                is_frac = True
            else:
                break
            pos += 1

        # TODO(scannell) need 'e'

        # TODO(scannell) This is awkward. Fix.
        self._pos = pos
        self.consume_all_space()

        if is_frac:
            return float(num) / divisor

        return num

    def consume_string(self):
        """Consume string.

        Returns:
          This returns a string or None if one isn't available.
        """
        spos = self._pos
        flag = None

        if spos == len(self._str):
            return None

        if self._str[spos] == 'b' or self._str[spos] == 'r':
            flag = self._str[spos]
            spos += 1

        if spos == len(self._str) or self._str[spos] not in ('"', '\''):
            return None

        if self._str[spos:spos + 3] == '"""':
            quotechar = self._str[spos:spos + 3]
        else:
            quotechar = self._str[spos]

        spos += len(quotechar)
        epos = spos
        while (epos < len(self._str) and
               self._str[epos:epos + len(quotechar)] != quotechar):
            epos += 1

        if epos == len(self._str):
            self.error('Expected string')

        self._pos = epos + len(quotechar)

        self.consume_all_space()

        return (self._str[spos:epos], flag)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#


from dataclasses import dataclass
from typing import Optional
from typing import List

from rfmt.blocks import LineBlock as LB
from rfmt.blocks import IndentBlock as IB
from rfmt.blocks import TextBlock as TB
from rfmt.blocks import StackBlock as SB
from rfmt.blocks import ChoiceBlock as CB
from rfmt.blocks import WrapBlock as WB

from .utils import with_commas
from .utils import comments_sqlf

from .const import SQLNumber
from .ident import SQLIdentifier
from .ident import SQLIdentifierPath
from .expr_op import SQLBiOp

from .node import SQLNode
from .node import SQLNodeList
from .expr import SQLExpr

from .query import SQLQuery
from .query import SQLAlias
from .query import SQLTableSource


@dataclass(frozen=True)
class SQLSetOp(SQLQuery):
    op: str
    left: SQLNode
    right: SQLNode

    def sqlf(self, compact):
        return SB(
            [self.left.sqlf(compact), TB(self.op),
             self.right.sqlf(compact)])


@dataclass(frozen=False)
class SQLField(SQLNode):
    expr: SQLNode
    alias: Optional[SQLAlias]
    comments: Optional[List[str]]

    @staticmethod
    def parse(lex) -> 'SQLField':
        expr = SQLExpr.parse(lex)
        alias = SQLAlias.consume(lex)
        return SQLField(expr, alias, lex.get_comments())

    def sqlf(self, compact):
        comments_sql = comments_sqlf(self.comments)

        compact_sql = self.expr.sqlf(True)
        if self.alias:
            compact_sql = LB(
                [compact_sql, TB(' '), self.alias.sqlf(True)])
        if comments_sql:
            compact_sql = SB(comments_sql + [compact_sql])

        if compact:
            return compact_sql

        full_sql = self.expr.sqlf(False)
        if self.alias:
            full_sql = SB(
                comments_sql + [full_sql,
                                IB(self.alias.sqlf(True))])
        elif comments_sql:
            full_sql = SB(comments_sql + [full_sql])

        return CB([compact_sql, full_sql])


@dataclass(frozen=True)
class SQLOrderedColumn(SQLNode):
    col: SQLNode
    dir: Optional[str] = None

    def sqlf(self, compact):
        if self.dir:
            return LB(
                [self.col.sqlf(compact), TB(' ' + self.dir)])
        return self.col.sqlf(compact)


@dataclass(frozen=True)
class SQLOrderLimitOffset(SQLNode):
    order: SQLNodeList[SQLOrderedColumn]
    limit: Optional[SQLNumber] = None
    offset: Optional[SQLNumber] = None

    def sqlf(self, compact):
        lines = []
        if self.order:
            lines.append(TB(' ORDER BY '))
            lines.append(LB(with_commas(compact, self.order)))
        if self.limit:
            lines.append(TB(' LIMIT '))
            lines.append(TB(str(self.limit)))
        if self.offset:
            lines.append(TB(' OFFSET '))
            lines.append(TB(str(self.offset)))
        compact_sql = LB(lines)

        if compact:
            return compact_sql

        stack = []
        if self.order:
            stack.append(TB(' ORDER BY'))
            stack.append(
                IB(WB(with_commas(compact, self.order))))
        if self.limit:
            stack.append(TB(' LIMIT'))
            stack.append(IB(TB(str(self.limit))))
        if self.offset:
            stack.append(TB(' OFFSET'))
            stack.append(IB(TB(str(self.offset))))

        return CB([
            compact_sql,
            SB(stack)
        ])

    @staticmethod
    def consume(lex) -> 'Optional[SQLOrderLimitOffset]':
        orders = []
        if lex.consume('ORDER'):
            lex.expect('BY')
            while True:
                oexpr = SQLExpr.parse(lex)
                if lex.consume('ASC'):
                    orders.append(SQLOrderedColumn(oexpr, 'ASC'))
                elif lex.consume('DESC'):
                    orders.append(SQLOrderedColumn(oexpr, 'DESC'))
                else:
                    orders.append(SQLOrderedColumn(oexpr, None))
                if not lex.consume(','):
                    break

        limit = None
        offset = None
        if lex.consume('LIMIT'):
            limit = (SQLNumber.consume(lex) or
                     lex.error('Expected limit number'))
            if lex.consume('OFFSET'):
                offset = (SQLNumber.consume(lex) or
                          lex.error('Expected offset number'))

        if orders or limit or offset:
            return SQLOrderLimitOffset(SQLNodeList(orders), limit, offset)

        return None


@dataclass(frozen=True)
class SQLOrderedQuery(SQLQuery):
    query: SQLQuery
    ordering: SQLNode

    def sqlf(self, compact):
        # TODO(scannell) Wrap in braces if it's not a SQLSelect type

        compact_sql = LB(
            [self.query.sqlf(True), self.ordering.sqlf(True)])
        if compact:
            return compact_sql

        return CB([
            compact_sql,
            SB([
                self.query.sqlf(compact),
                self.ordering.sqlf(compact),
            ])
        ])


@dataclass(frozen=False)
class SQLSelect(SQLQuery):
    fields: SQLNodeList[SQLField]
    where_expr: Optional[SQLExpr] = None
    groups: Optional[SQLNodeList[SQLField]] = None
    from_tables: Optional[SQLNode] = None
    select_as_type: Optional[str] = None
    select_type: Optional[str] = None
    having_expr: Optional[SQLExpr] = None
    qualify_expr: Optional[SQLExpr] = None
    comments: Optional[List[str]] = None

    def sqlf(self, compact):

        select_clause = [TB('SELECT')]
        if self.select_as_type:
            select_clause.append(TB(' AS ' + self.select_as_type))
        if self.select_type:
            select_clause.append(TB(' ' + self.select_type))

        lines = []
        lines.append(LB(select_clause))
        lines.append(TB(' '))
        lines.append(LB(with_commas(True, self.fields)))
        if self.from_tables:
            lines.append(TB(' FROM '))
            lines.append(self.from_tables.sqlf(True))
        if self.where_expr:
            lines.append(TB(' WHERE '))
            lines.append(self.where_expr.sqlf(True))
        if self.groups:
            lines.append(TB(' GROUP BY '))
            lines.append(LB(with_commas(True, self.groups)))
        if self.having_expr:
            lines.append(TB(' HAVING '))
            lines.append(self.having_expr.sqlf(True))
        if self.qualify_expr:
            lines.append(TB(' QUALIFY '))
            lines.append(self.qualify_expr.sqlf(True))
        compact_sql = LB(lines)

        if compact:
            return compact_sql

        stack = comments_sqlf(self.comments)
        stack.append(LB(select_clause))
        stack.append(IB(SB(with_commas(compact, self.fields))))
        if self.from_tables:
            stack.append(TB('FROM'))
            stack.append(IB(self.from_tables.sqlf(compact)))
        if self.where_expr:
            stack.append(TB('WHERE'))
            stack.append(IB(self.where_expr.sqlf(compact)))
        if self.groups:
            stack.append(TB('GROUP BY'))
            stack.append(
                IB(WB(with_commas(compact, self.groups))))
        if self.having_expr:
            stack.append(TB('HAVING'))
            stack.append(IB(self.having_expr.sqlf(compact)))
        if self.qualify_expr:
            stack.append(TB('QUALIFY'))
            stack.append(IB(self.qualify_expr.sqlf(compact)))

        return SB(stack)

    @staticmethod
    def parse(lex) -> 'SQLSelect':

        comments = lex.get_comments()

        lex.expect('SELECT')

        select_as_type = None
        if lex.consume('AS'):
            select_as_type = (lex.consume('STRUCT') or
                              lex.consume('VALUE') or
                              lex.error('Expected STRUCT or VALUE'))

        select_type = (lex.consume('DISTINCT') or
                       lex.consume('ALL') or
                       None)

        if not select_as_type and lex.consume('AS'):
            select_as_type = (lex.consume('STRUCT') or
                              lex.consume('VALUE') or
                              lex.error('Expected STRUCT or VALUE'))
        
        from_consumed = False
        fields: List[SQLField] = []
        while True:
            fields.append(SQLField.parse(lex))
            if not lex.consume(','):
                break
            from_consumed = lex.consume('FROM')
            if from_consumed:
                break

        from_tables = (from_consumed or lex.consume('FROM')) and SQLFrom.parse(lex)
        where_expr = lex.consume('WHERE') and SQLExpr.parse(lex)

        groups: List[SQLField] = []
        if lex.consume('GROUP'):
            lex.expect('BY')
            while True:
                groups.append(SQLField.parse(lex))
                if not lex.consume(','):
                    break

        having_expr = lex.consume('HAVING') and SQLExpr.parse(lex)
        qualify_expr = lex.consume('QUALIFY') and SQLExpr.parse(lex)

        # TODO(scannell) Support WINDOW
        # Error on Window
        if lex.consume('WINDOW'):
            lex.error('WINDOW unsupported')

        comments.extend(lex.get_comments())

        return SQLSelect(SQLNodeList(fields), where_expr,
                         SQLNodeList(groups), from_tables,
                         select_as_type, select_type, having_expr,
                         qualify_expr, comments)


@dataclass(frozen=True)
class SQLSubSelect(SQLQuery):
    query: SQLQuery
    alias: Optional[SQLAlias] = None

    def sqlf(self, compact):
        stack = [
            TB('('), IB(self.query.sqlf(compact)),
            TB(')')
        ]
        if self.alias:
            stack.append(LB([TB(' '), self.alias.sqlf(True)]))
        return SB(stack)


@dataclass(frozen=True)
class SQLWithSelect(SQLQuery):
    tables: SQLNodeList[SQLIdentifier]
    sqls: SQLNodeList[SQLQuery]
    select: SQLNode

    def sqlf(self, compact):
        if not self.tables:
            return self.select.sqlf(compact)
        stack = [TB('WITH')]
        for i, table in enumerate(self.tables):
            iblock = [LB([table.sqlf(True), TB(' AS (')])]
            iblock.append(IB(self.sqls[i].sqlf(compact)))
            if i == len(self.tables) - 1:
                iblock.append(TB(')'))
            else:
                iblock.append(TB('),'))
            stack.append(IB(SB(iblock)))

        stack.append(self.select.sqlf(compact))

        return SB(stack)

    @staticmethod
    def parse(lex) -> 'SQLWithSelect':
        return (SQLWithSelect.consume(lex) or
                lex.error('Expected SELECT statement'))

    @staticmethod
    def consume(lex) -> 'Optional[SQLWithSelect]':
        tables: List[SQLIdentifier] = []
        sqls: List[SQLQuery] = []
        if lex.consume('WITH'):
            while True:
                tables.append(SQLIdentifier.parse(lex))
                lex.expect('AS')
                lex.expect('(')
                sqls.append(SQLQuery.parse(lex))
                lex.expect(')')
                if not lex.consume(','):
                    break

        query: SQLQuery
        if lex.peek('SELECT'):
            query = SQLSelect.parse(lex)
        elif not tables and (lex.peek(['(', 'WITH']) or
                             lex.peek(['(', 'SELECT'])):
            lex.consume('(')
            query = SQLSubSelect(SQLQuery.parse(lex), None)
            lex.expect(')')
        elif tables:
            lex.error('Expected SELECT or sub-query')
        else:
            return None

        order_limit_offset = SQLOrderLimitOffset.consume(lex)

        if order_limit_offset:
            query = SQLOrderedQuery(query, order_limit_offset)

        # These need to create a function that binds these things together
        if lex.consume('INTERSECT'):
            lex.expect('DISTINCT')
            query = SQLSetOp('INTERSECT DISTINCT', query,
                             SQLQuery.parse(lex))

        elif lex.consume('EXCEPT'):
            lex.expect('DISTINCT')
            query = SQLSetOp('EXCEPT DISTINCT', query,
                             SQLQuery.parse(lex))

        elif lex.consume('MINUS'):
            query = SQLSetOp('MINUS', query,
                             SQLQuery.parse(lex))

        elif lex.consume('UNION'):
            union_op = 'UNION'
            if lex.consume('ALL'):
                union_op += ' ALL'
            elif lex.consume('DISTINCT'):
                union_op += ' DISTINCT'
            query = SQLSetOp(union_op, query, SQLQuery.parse(lex))

        # Return query with orders, limit, bindings, etc
        return SQLWithSelect(SQLNodeList(tables), SQLNodeList(sqls), query)


@dataclass(frozen=True)
class SQLFrom(SQLNode):
    base: SQLNode
    joins: SQLNodeList[SQLNode]

    def sqlf(self, compact):
        fields = [self.base.sqlf(compact)]

        for j in self.joins:
            if compact:
                fields.append(TB(' '))
            fields.append(j.sqlf(compact))

        if compact:
            return LB(fields)

        return SB(fields)

    @staticmethod
    def parse(lex) -> 'SQLFrom':
        base_table = SQLTableSource.parse(lex)

        joins = []
        prev_alias = base_table.alias
        while True:
            join = SQLJoin.parse(lex)
            if not join:
                break
            if join.join_keyword == 'USING':
                join.fix_left_alias(prev_alias)
                prev_alias = join.table.alias
            joins.append(join)

        return SQLFrom(base_table, SQLNodeList(joins))


@dataclass(frozen=True)
class SQLJoin(SQLNode):
    join_type: str
    table: SQLNode
    join_expr: Optional[SQLNode]
    join_keyword: Optional[str]

    def sqlf(self, compact):
        join_type = TB(self.join_type + ' JOIN ')
        if self.join_expr is None:
            return LB([join_type, self.table.sqlf(compact)])
        compact_sql = LB([
            join_type, self.table.sqlf(True), TB(' ON '),
            self.join_expr.sqlf(True)
        ])
        if compact:
            return compact_sql
        return CB([
            compact_sql, SB([
                SB([
                    join_type,
                    self.table.sqlf(False),
                    TB(' ON')
                ]),
                IB(self.join_expr.sqlf(False))
            ])
        ])

    @staticmethod
    def parse_type(lex) -> Optional[str]:

        join_type = lex.consume_any(['INNER', 'CROSS'])
        if join_type:
            lex.expect('JOIN')
            return join_type

        join_type = lex.consume_any(['FULL', 'LEFT', 'RIGHT'])
        if join_type:
            if lex.consume('OUTER'):
                join_type = join_type + ' OUTER'
            lex.expect('JOIN')
            return join_type

        if lex.consume('JOIN'):
            return 'INNER'

        if lex.consume(','):
            return 'CROSS'

        return None

    @staticmethod
    def parse(lex) -> 'Optional[SQLJoin]':
        join_type = SQLJoin.parse_type(lex)
        if join_type is None:
            return None
        join_table = SQLTableSource.parse(lex)
        join_expr = None
        join_keyword = None
        if lex.consume('ON'):
            join_expr = SQLExpr.parse(lex)
            join_keyword = 'ON'
        elif lex.consume('USING'):
            lex.expect('(')
            join_expr = None
            while True:
                field = SQLIdentifierPath.parse(lex)
                if join_table.alias is None:
                    join_table.alias = SQLAlias(SQLIdentifier(join_table.table.names[-1].value))
                right = SQLIdentifierPath(SQLNodeList([join_table.alias.alias]) + field.names)
                if join_expr is None: 
                    join_expr = SQLBiOp('=', field, right)
                else:
                    join_expr = SQLBiOp('AND', join_expr, SQLBiOp('=', field, right))
                if lex.consume(')'):
                    break
                lex.consume(',')
            join_keyword = 'USING'
        return SQLJoin(join_type, join_table, join_expr, join_keyword)

    def fix_left_alias(self, alias: SQLAlias):
        if self.join_keyword != 'USING':
            return
        expr = self.join_expr
        while True:
            if expr.sql_op == '=':
                left = expr.left
                names = left.names   
                left.names = SQLNodeList([alias.alias]) + names
                break
            elif expr.sql_op == 'AND':
                right_expr = expr.right
                left = right_expr.left
                names = left.names   
                left.names = SQLNodeList([alias.alias]) + names
                expr = expr.left


# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#

from dataclasses import dataclass
from typing import List
from typing import Optional

from rfmt.blocks import ChoiceBlock as CB
from rfmt.blocks import IndentBlock as IB
from rfmt.blocks import LineBlock as LB
from rfmt.blocks import StackBlock as SB
from rfmt.blocks import TextBlock as TB
from rfmt.blocks import WrapBlock as WB

from .utils import with_commas

from .query import SQLQuery
from .query_impl import SQLOrderLimitOffset
from .expr import SQLExpr
from .const import SQLConstant
from .const import SQLNumber
from .ident import SQLIdentifierPath
from .node import SQLNode
from .node import SQLNodeList
from .expr_funcs import SQLFuncExpr, SQLCustomFuncs


@dataclass(frozen=True)
class SQLArrayLiteral(SQLExpr):
    args: SQLNodeList

    def sqlf(self, compact):
        compact_sql = LB([
            TB('['),
            LB(with_commas(True, self.args)),
            TB(']')
        ])
        if compact:
            return compact_sql
        return CB([
            compact_sql,
            LB([
                TB('['),
                WB(with_commas(compact, self.args, tail=']'))
            ]),
        ])

    @staticmethod
    def consume(lex) -> 'Optional[SQLArrayLiteral]':
        if not lex.consume('['):
            return None
        exprs: List[SQLExpr] = []
        while True:
            exprs.append(SQLExpr.parse(lex))
            if not lex.consume(','):
                break
        lex.expect(']')
        return SQLArrayLiteral(SQLNodeList(exprs))


@dataclass(frozen=True)
class SQLArraySelect(SQLExpr):
    query: SQLQuery

    def sqlf(self, compact):
        compact_sql = LB([
            TB('ARRAY('), self.query.sqlf(True), TB(')')
        ])

        if compact:
            return compact_sql

        return CB([
            compact_sql,
            SB([
                TB('ARRAY('),
                IB(self.query.sqlf(compact)),
                TB(')')
            ])
        ])

    @staticmethod
    def consume(lex) -> 'Optional[SQLArraySelect]':
        if not lex.consume('ARRAY'):
            return None

        lex.expect('(')
        query = SQLQuery.parse(lex)
        lex.expect(')')
        return SQLArraySelect(query)


@dataclass(frozen=True)
class SQLArrayAgg(SQLExpr):
    is_distinct: bool
    expr: SQLNode
    nulls: Optional[str]
    order_limit_offset: Optional[SQLOrderLimitOffset]
    analytic: Optional[SQLNode]
    offset: Optional[int]

    def sqlf(self, compact):
        lines = [TB('ARRAY_AGG(')]
        if self.is_distinct:
            lines.append(TB('DISTINCT '))
        lines.append(self.expr.sqlf(True))
        if self.nulls:
            lines.append(TB(self.nulls) + ' NULLS')
        if self.order_limit_offset:
            lines.append(self.order_limit_offset.sqlf(True))
        if self.analytic:
            lines.append(self.analytic.sqlf(True))
        lines.append(TB(')'))
        if self.offset:
            lines.append(TB('[OFFSET('))
            lines.append(TB(' ') )
            lines.append(self.offset.sqlf(compact))
            lines.append(TB(')]'))

        compact_sql = LB(lines)

        if compact:
            return compact_sql

        stack = [TB('ARRAY_AGG(')]
        indent = []
        if self.is_distinct:
            indent.append(
                LB([TB('DISTINCT '), self.expr.sqlf(compact)]))
        else:
            indent.append(self.expr.sqlf(compact))
        if self.nulls:
            indent.append(TB(self.nulls) + ' NULLS')
        if self.order_limit_offset:
            indent.append(self.order_limit_offset.sqlf(compact))
        if self.analytic:
            indent.append(self.analytic.sqlf(compact))
        stack.append(IB(SB(indent)))
        stack.append(TB(')'))
        if self.offset:
            stack.append(TB('[OFFSET('))
            stack.append(TB(' ') )
            stack.append(self.offset.sqlf(compact))
            stack.append(TB(')]'))

        return CB([
            compact_sql,
            SB(stack)
        ])

    @staticmethod
    def consume(lex) -> 'Optional[SQLArrayAgg]':
        if not lex.consume('ARRAY_AGG'):
            return None

        lex.expect('(')

        is_distinct = bool(lex.consume('DISTINCT'))

        expr = SQLExpr.parse(lex)

        nulls = None
        if lex.consume('IGNORE'):
            nulls = 'IGNORE'
            lex.expect('NULLS')
        elif lex.consume('RESPECT'):
            nulls = 'RESPECT'
            lex.expect('NULLS')

        order_limit_offset = SQLOrderLimitOffset.consume(lex)

        analytic = SQLAnalytic.consume(lex)

        lex.expect(')')

        offset = None
        if lex.consume('['):
            lex.expect('OFFSET')
            lex.consume('(')
            offset = SQLNumber.parse(lex)
            lex.consume(')')
            lex.expect(']')

        return SQLArrayAgg(is_distinct, expr, nulls,
                           order_limit_offset, analytic, offset)


@dataclass(frozen=True)
class SQLStringAgg(SQLExpr):
    name: str
    is_distinct: bool
    expr: SQLNode
    delimiter: Optional[str]
    nulls: Optional[str]
    order_limit_offset: Optional[SQLOrderLimitOffset]
    analytic: Optional[SQLNode]
    analytic_name: Optional[str]
    number: Optional[int]

    def sqlf(self, compact):
        lines = [TB('{}('.format(self.name))]
        if self.is_distinct:
            lines.append(TB('DISTINCT '))
        lines.append(self.expr.sqlf(True))
        if self.delimiter:
            lines.append(TB(', '+ self.delimiter.as_sql(compact)))
        if self.nulls:
            lines.append(TB(self.nulls) + ' NULLS')
        if self.order_limit_offset:
            lines.append(self.order_limit_offset.sqlf(True))
        if self.analytic:
            lines.append(self.analytic.sqlf(True))
        lines.append(TB(')'))
        if self.number:
            if self.analytic_name:
                lines.append(TB('[{}('.format(self.analytic_name)))
                lines.append(TB(' ') )
                lines.append(self.number.sqlf(compact))
                lines.append(TB(')]'))
            else:
                lines.append(TB('['))
                lines.append(self.number.sqlf(compact))
                lines.append(TB(']'))

        compact_sql = LB(lines)

        if compact:
            return compact_sql

        stack = [TB('{}('.format(self.name))]
        indent = []
        if self.is_distinct:
            args = [TB('DISTINCT '), self.expr.sqlf(compact)]
        else:
            args = [self.expr.sqlf(compact)]
        if self.delimiter:
            args.append(TB(', ' + self.delimiter.as_sql(compact)))
        indent.append(LB(args))
        
        if self.nulls:
            indent.append(TB(self.nulls) + ' NULLS')
        if self.order_limit_offset:
            indent.append(self.order_limit_offset.sqlf(compact))
        if self.analytic:
            indent.append(self.analytic.sqlf(compact))
        stack.append(IB(SB(indent)))
        stack.append(TB(')'))
        if self.number:
            if self.analytic_name:
                stack.append(TB('[{}('.format(self.analytic_name)))
                stack.append(TB(' ') )
                stack.append(self.number.sqlf(compact))
                stack.append(TB(')]'))
            else:
                stack.append(TB('['))
                stack.append(self.number.sqlf(compact))
                stack.append(TB(']'))

        return CB([
            compact_sql,
            SB(stack)
        ])

    @staticmethod
    def consume(lex) -> 'Optional[SQLStringAgg]':
        name = lex.consume('STRING_AGG') or lex.consume('SPLIT')
        if not name:
            return None

        lex.expect('(')

        is_distinct = bool(lex.consume('DISTINCT'))

        expr = SQLExpr.parse(lex)
        delimiter = None

        if lex.consume(','):
            delimiter = SQLConstant.consume(lex)

        nulls = None
        if lex.consume('IGNORE'):
            nulls = 'IGNORE'
            lex.expect('NULLS')
        elif lex.consume('RESPECT'):
            nulls = 'RESPECT'
            lex.expect('NULLS')

        order_limit_offset = SQLOrderLimitOffset.consume(lex)

        analytic = SQLAnalytic.consume(lex)

        lex.expect(')')

        analytic_name = None
        number = None
        if lex.consume('['):
            analytic_name = SQLConstant.consume(lex)
            lex.consume('(')
            number = SQLNumber.parse(lex)
            lex.consume(')')
            lex.expect(']')

        return SQLStringAgg(name, is_distinct, expr, delimiter, nulls,
                           order_limit_offset, analytic, analytic_name, number)


@dataclass(frozen=True)
class SQLExprWithAnalytic(SQLExpr):
    function: SQLExpr
    analytic: SQLNode

    def sqlf(self, compact):
        compact_sqlf = LB([self.function.sqlf(compact), TB(' '),
                           self.analytic.sqlf(compact)])
        if compact:
            return compact_sqlf

        return CB([
            compact_sqlf,
            SB([self.function.sqlf(compact),
                self.analytic.sqlf(compact)])
        ])

    @staticmethod
    def parse(lex) -> 'SQLExpr':

        # Try alternatives first
        expr: SQLExpr = (SQLConstant.consume(lex) or
                         SQLArrayLiteral.consume(lex) or
                         SQLArrayAgg.consume(lex) or
                         SQLStringAgg.consume(lex) or
                         SQLArraySelect.consume(lex) or
                         SQLCustomFuncs.consume(lex) or
                         SQLIdentifierPath.parse(lex))

        # If it is a SQLIdentifierPath, it may be
        # a normal function call.
        if isinstance(expr, SQLIdentifierPath) and lex.consume('('):

            # Parse as a function
            func_args: List[SQLExpr] = []
            if not lex.consume(')'):
                while True:
                    func_args.append(SQLExpr.parse(lex))
                    if not lex.consume(','):
                        break
                lex.expect(')')

            # Turn it into a function
            expr = SQLFuncExpr(expr.names, SQLNodeList(func_args))

        window = SQLAnalytic.consume(lex)
        if window:
            expr = SQLExprWithAnalytic(expr, window)

        return expr


@dataclass(frozen=True)
class SQLAnalytic(SQLExpr):
    partition_by: SQLNodeList
    order_by: SQLNodeList
    range_desc: str

    def sqlf(self, compact):
        lines = []
        lines.append(TB('OVER ('))
        if self.partition_by:
            lines.append(TB('PARTITION BY '))
            lines.extend(with_commas(True, self.partition_by, ','))
        if self.order_by:
            lines.append(TB(' ORDER BY '))
            lines.extend(with_commas(True, self.order_by, ','))
        if self.range_desc:
            lines.append(TB(self.range_desc))
        lines.append(TB(')'))

        if compact:
            return LB(lines)

        full_sql = [
            TB('OVER ('),
        ]
        if self.partition_by:
            full_sql.append(
                IB(
                    SB([
                        TB('PARTITION BY'), IB(
                            WB(with_commas(True, self.partition_by)))
                    ])))
        if self.order_by:
            full_sql.append(
                IB(
                    SB([
                        TB('ORDER BY'), IB(
                            WB(with_commas(True, self.order_by)))
                    ])))
        if self.range_desc:
            full_sql.append(IB(TB(self.range_desc)))
        full_sql.append(TB(')'))

        r = CB([LB(lines), SB(full_sql)])
        return r

    @staticmethod
    def consume(lex) -> 'Optional[SQLAnalytic]':
        if not lex.consume('OVER'):
            return None

        lex.expect('(')

        partition_by = []
        if lex.consume('PARTITION'):
            lex.expect('BY')
            while True:
                partition_by.append(SQLExpr.parse(lex))
                if not lex.consume(','):
                    break

        order_by = []
        if lex.consume('ORDER'):
            lex.expect('BY')
            while True:
                oby = SQLExpr.parse(lex)
                order = None
                if lex.consume('ASC'):
                    order = 'ASC'
                elif lex.consume('DESC'):
                    order = 'DESC'
                # TODO(scannell): Capture ASC/DESC
                order_by.append(oby)
                if not lex.consume(','):
                    break

        win_spec = (lex.consume('ROWS') or lex.consume('RANGE'))

        # If window specified, parse it out
        if win_spec:
            if lex.consume('BETWEEN'):
                win_spec += (' BETWEEN ' +
                             SQLAnalytic._parse_frame_boundary(lex))
                lex.expect('AND')
                win_spec += ' AND ' + SQLAnalytic._parse_frame_boundary(lex)
            else:
                win_spec += ' ' + SQLAnalytic._parse_frame_boundary(lex)

        lex.expect(')')

        # Return analytics function
        return SQLAnalytic(SQLNodeList(partition_by),
                           SQLNodeList(order_by),
                           win_spec)

    @staticmethod
    def _parse_frame_boundary(lex):
        if lex.consume('UNBOUNDED'):
            typ = lex.consume('PRECEDING') or lex.consume('FOLLOWING') or lex.error('Expected PRECEDING or FOLLOWING')
            return 'UNBOUNDED {}'.format(typ)

        if lex.consume('CURRENT'):
            lex.expect('ROW')
            return 'CURRENT ROW'

        num = SQLNumber.consume(lex)

        num_typ = (lex.consume('PRECEDING') or lex.consume('FOLLOWING') or
                   lex.error('Expected PRECEDING or FOLLOWING'))

        return '{} {}'.format(num, num_typ)



#  Copyright 2015 Google Inc. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Base functionality for the R formatter.

- Defines formatter-specific exception class.
- Access to and manipulation of tool options.
"""


class Error(Exception):
  """Base class for R formatter exceptions."""

  def __init__(self, msg):
    super(Error, self).__init__()
    self.msg = msg

  def __str__(self):
    return self.msg


class Options(object):
  """Singleton class holding global options for the formatter."""

  _instance = None

  def __new__(cls):
    if Options._instance is None:
      Options._instance = super(Options, cls).__new__(cls)
    return cls._instance

  def Set(self, flags=None, **opts):
    """Set the options used throughout the rfmt system.

    Args:
      flags: an optional :-) Google-style (pyglib) FLAGS object.
      **opts: other keyword options.
    """
    if flags:
      self.__dict__.update(flags.FlagValuesDict())
    self.__dict__.update(opts)
    self.Check()

  def Check(self):
    """Assertion verification for options."""
    try:
      assert self.m0 >= 0, "margin0"
      assert self.m1 >= self.m0, "margin1"
      assert self.c0 >= 0, "cost0"
      assert self.c1 >= 0, "cost1"
      assert self.cb >= 0, "costb"
      assert self.ind >= 0, "indent"
      assert self.adj_comment >= 0, "adj_comment"
      assert self.adj_flow >= 0, "adj_flow"
      assert self.adj_call >= 0, "adj_call"
      assert self.adj_arg >= 0, "adj_arg"
      assert self.cpack >= 0, "cpack"
    except AssertionError as e:
      raise Error("Illegal option value for '%s'" % e.args[0])




# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#

import unittest

from .lexer import SQLLexer
from .lexer import ParsingError


class TestLexer(unittest.TestCase):

    def test_end(self):

        # no expection
        SQLLexer('').expect_end()

        # expection
        with self.assertRaises(ParsingError):
            SQLLexer('not end').expect_end()

        # no expection
        SQLLexer('').expect_end()

    def check_expect(self, lex, elem, res):

        # Check basic
        self.assertEqual(SQLLexer(lex).consume(elem), res)

        # If we're expect None, then let's ensure expect works, too
        if res is None:
            with self.assertRaises(ParsingError):
                SQLLexer(lex).expect(elem)

    def test_elements(self):
        self.check_expect('+', '+', '+')
        self.check_expect('+', '/', None)

        self.check_expect('if', 'IF', 'IF')
        self.check_expect('ifs', 'IF', None)
        self.check_expect('if2', 'IF', None)
        self.check_expect('if(', 'IF', 'IF')
        self.check_expect('+', 'IF', None)
        self.check_expect('i', 'IF', None)

    def check_identifier(self, lex, rep):

        # Pull out the identifier
        val = SQLLexer(lex).consume_identifier()

        # Same..
        self.assertEqual(val, rep)

    def test_identifier(self):
        self.check_identifier('thisIsIt', 'thisIsIt')
        self.check_identifier('`this IsIt`', 'this IsIt')
        self.check_identifier('myID+', 'myID')
        self.check_identifier('', None)
        self.check_identifier('2', None)
        self.check_identifier('ASSERT_ROWS_MODIFIED', None)

    def check_const(self, lex, sql):
        lex = SQLLexer(lex)
        val = lex.consume_identifier()
        self.assertEqual(val, sql)
        lex.consume_all_space()
        self.assertEqual(lex.consume_identifier(), None)

    def test_sqlcomment(self):
        self.check_const('  -- this is all space\nIdentifier', 'Identifier')
        self.check_const(
            '  -- this is all space\nAgain -- end of file\n -- again',
            'Again')

    def test_ccomment(self):
        self.check_const('  /* this is all space\n asdf */ \n varTable',
                         'varTable')

    def test_hashcomment(self):
        self.check_const('  /* this is all space\n asdf */ \n # comment \n'
                         'f10 # end of file',
                         'f10')

    def check_comments(self, lex, comments, sql):
        lex = SQLLexer(lex)
        val = lex.consume_identifier()
        self.assertEqual(val, sql)
        self.assertEqual(comments, lex.get_comments())
        lex.consume_all_space()
        self.assertEqual(lex.consume_identifier(), None)

    def test_comment_extract(self):
        self.check_comments('  -- comment 1\n  # comment 2..\n  /* okay */ '
                            'Identifier',
                            [u'comment 1', u'okay'],
                            'Identifier')



#  Copyright 2015 Google Inc. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""The block language system for the R language formatter."""

from . import base
from . import support
import re

_options = base.Options()  # Shorthand for convenient access


class LayoutBlock(object):
  """The abstract class at base of the block hierarchy."""

  def __init__(self, is_breaking=False):
    # If a newline is mandated after this block.
    self.is_breaking = is_breaking
    # See OptLayout method below for use of layout_cache.
    self.layout_cache = {}

  def Parms(self):
    """A dictionary containing the parameters of this block."""
    return {}

  def ReprParms(self):
    """The printed representation of this block's parameters."""
    if not self.Parms(): return ''
    return ('<%s>' % (', '.join('%s=%s' % (key, val.__repr__())
                                for key, val in self.Parms().iteritems())))

  def __repr__(self):
    return re.sub('[a-z]', '', self.__class__.__name__ +
                  '*' * self.is_breaking) + self.ReprParms()

  def OptLayout(self, rest_of_line):
    """Retrieve or compute the least-cost (optimum) layout for this block.

    Args:
      rest_of_line: a Solution object representing the text to the right of
        this block.
    Returns:
      A Solution object representing the optimal layout for this block and
      the rest of the line.
    """
    # Deeply-nested choice block may result in the same continuation supplied
    # repeatedly to the same block. Without memoisation, this may result in an
    # exponential blow-up in the layout algorithm.
    if rest_of_line not in self.layout_cache:
      self.layout_cache[rest_of_line] = self.DoOptLayout(rest_of_line)
    return self.layout_cache[rest_of_line]

  def DoOptLayout(self, rest_of_line):
    """Compute the least-cost (optimum) layout for this block.

    Args:
      rest_of_line: a Solution object representing the text to the right of
        this block.
    Returns:
      A Solution object representing the optimal layout for this block and
      the rest of the line.
    """
    # Abstract method.
    pass

  def PrintOn(self, outp):
    """Print the contents of this block with the optimal layout.

    Args:
      outp: a stream on which output is to be printed.
    """
    soln = self.OptLayout(None)
    support.Console(outp).PrintLayout(soln.layouts[0])


class TextBlock(LayoutBlock):
  """A block containing a single unbroken string."""

  def __init__(self, text, is_breaking=False):
    super(TextBlock, self).__init__(is_breaking)
    self.text = text

  def __repr__(self):
    return '*' * self.is_breaking + self.text

  def DoOptLayout(self, rest_of_line):
    span = len(self.text)
    layout = support.Layout([support.LayoutElement.String(self.text)])
    # The costs associated with the layout of this block may require 1, 2 or 3
    # knots, depending on how the length of the text compares with the two
    # margins (m0 and m1) in _options. Note that we assume
    # _options.m1 >= _options.m0 >= 0, as asserted in base.Options.Check().
    if span >= _options.m1:
      s = support.Solution([0], [span],
                           [(span - _options.m0) * _options.c0 +
                            (span - _options.m1) * _options.m1],
                           [_options.c0 + _options.c1], [layout])
    elif span >= _options.m0:
      s = support.Solution([0, _options.m1 - span], [span] * 2,
                           [(span - _options.m0) * _options.c0,
                            (_options.m1 - _options.m0) * _options.c0],
                           [_options.c0, _options.c0 + _options.c1],
                           [layout] * 2)
    else:
      s = support.Solution([0, _options.m0 - span, _options.m1 - span],
                           [span] * 3,
                           [0, 0, (_options.m1 - _options.m0) * _options.c0],
                           [0, _options.c0, _options.c0 + _options.c1],
                           [layout] * 3)
    return s.WithRestOfLine(rest_of_line)


class CompositeLayoutBlock(LayoutBlock):
  """The abstract superclass of blocks which contain other blocks (elements).

  Note that we assume at least one element.
  """

  def __init__(self, elements):
    super(CompositeLayoutBlock, self).__init__()
    self.elements = elements
    # Break after this block if its last element requires a break.
    self.is_breaking = elements and elements[-1].is_breaking

  def ReprElements(self):
    return '[%s]' % (', '.join(e.__repr__() for e in self.elements))

  def __repr__(self):
    return super(CompositeLayoutBlock, self).__repr__() + self.ReprElements()


class LineBlock(CompositeLayoutBlock):
  """A block that places its elements in a single line."""

  def __init__(self, elements):
    super(LineBlock, self).__init__(elements)

  def DoOptLayout(self, rest_of_line):
    if not self.elements: return rest_of_line
    element_lines = [[]]
    for i, elt in enumerate(self.elements):
      element_lines[-1].append(elt)
      if i < len(self.elements) - 1 and elt.is_breaking:
        element_lines.append([])
    if len(element_lines) > 1:
      element_lines = _options.format_policy.BreakElementLines(element_lines)
    line_solns = []
    for i, ln in enumerate(element_lines):
      ln_layout = None if i < len(element_lines) - 1 else rest_of_line
      for elt in ln[::-1]:
        ln_layout = elt.OptLayout(ln_layout)
      line_solns.append(ln_layout)
    soln = support.VSumSolution(line_solns)
    return soln.PlusConst(_options.cb * (len(line_solns) - 1))


def IndentBlock(element, indent=None):
  """Return a block that contains another block, indented by a given amount."""
  if indent is None: indent = _options.ind
  return LineBlock([TextBlock(' ' * indent), element])


class ChoiceBlock(CompositeLayoutBlock):
  """A block which contains alternate layouts of the same content."""

  # Note: All elements of a ChoiceBlock are breaking, if any are.
  def __init__(self, elements):
    super(ChoiceBlock, self).__init__(elements)

  def DoOptLayout(self, rest_of_line):
    # The optimum layout of this block is simply the piecewise minimum of its
    # elements' layouts.
    return support.MinSolution([e.OptLayout(rest_of_line)
                                for e in self.elements])


class MultBreakBlock(CompositeLayoutBlock):
  """The abstract superclass of blocks that locally modify line break cost."""

  def __init__(self, elements, break_mult=1):
    super(MultBreakBlock, self).__init__(elements)
    self.break_mult = break_mult

  def Parms(self):
    return dict(super(MultBreakBlock, self).Parms().items() +
                [('break_mult', self.break_mult)])


class StackBlock(MultBreakBlock):
  """A block that arranges its elements vertically, separated by line breaks."""

  def __init__(self, elements, break_mult=1):
    super(StackBlock, self).__init__(elements, break_mult)

  def DoOptLayout(self, rest_of_line):
    # The optimum layout for this block arranges the elements vertically. Only
    # the final element is composed with the continuation provided---all the
    # others see an empty continuation ("None"), since they face the end of
    # a line.
    if not self.elements: return rest_of_line
    soln = support.VSumSolution([e.OptLayout(None)
                                 for e in self.elements[:-1]] +
                                [self.elements[-1].OptLayout(rest_of_line)])
    # Under some odd circumstances involving comments, we may have a degenerate
    # solution.
    if soln is None:
      return rest_of_line
    # Add the cost of the line breaks between the elements.
    return soln.PlusConst(_options.cb * self.break_mult *
                          max(len(self.elements) - 1, 0))


class WrapBlock(MultBreakBlock):
  """A block that arranges its elements like a justified paragraph."""

  def __init__(self, elements, sep=' ', break_mult=1, prefix=None):
    super(WrapBlock, self).__init__(elements)
    self.break_mult = break_mult
    self.sep = sep
    self.prefix = prefix
    self.elt_is_breaking = [e.is_breaking for e in elements]
    self.n = len(self.elements)

  def Parms(self):
    return dict(super(WrapBlock, self).Parms().items() +
                [('sep', self.sep)] +
                (self.prefix is not None) * [('prefix', self.prefix)])

  def DoOptLayout(self, rest_of_line):
    # Computing the optimum layout for this class of block involves finding the
    # optimal packing of elements into lines, a problem which we address using
    # dynamic programming.
    sep_layout = TextBlock(self.sep).OptLayout(None)
    # TODO(pyelland): Investigate why OptLayout doesn't work here.
    prefix_layout = self.prefix and TextBlock(self.prefix).DoOptLayout(None)
    elt_layouts = [e.OptLayout(None) for e in self.elements]
    # Entry i in the list wrap_solutions contains the optimum layout for the
    # last n - i elements of the block.
    wrap_solutions = [None] * self.n
    # Note that we compute the entries for wrap_solutions in reverse order,
    # at each iteration considering all the elements from i ... n - 1 (the
    # actual number of elements considered increases by one on each iteration).
    # This means that the complete solution, with elements 0 ... n - 1 is
    # computed last.
    for i in range(self.n - 1, -1, -1):
      # To calculate wrap_solutions[i], consider breaking the last n - i
      # elements after element j, for j = i ... n - 1.
      # By induction, wrap_solutions contains the optimum layout of the
      # elements after the break, so the full layout is calculated by composing
      # a line with the elements before the break with the entry from
      # wrap_solutions corresponding to the elements after the break.
      # The optimum layout to be entered into wrap_solutions[i] is then simply
      # the minimum of the full layouts calculated for each j.
      solutions_i = []
      # The layout of the elements before the break is built up incrementally
      # in line_layout.
      if prefix_layout is None:
        line_layout = elt_layouts[i]
      else:
        line_layout = prefix_layout.WithRestOfLine(elt_layouts[i])
      last_breaking = self.elements[i].is_breaking
      for j in range(i, self.n - 1):
        full_soln = support.VSumSolution([line_layout, wrap_solutions[j + 1]])
        # We adjust the cost of the full solution by adding the cost of the
        # line break we've introduced, and a small penalty (_options.cpack) to
        # favor (ceteris paribus) layouts with elements packed into earlier
        # lines.
        solutions_i.append(full_soln.PlusConst(_options.cb * self.break_mult +
                                               _options.cpack * (self.n - j)))
        # If the element at the end of the line mandates a following line break,
        # we're done.
        if last_breaking: break
        # Otherwise, add a separator and the next element to the line layout
        # and continue.
        sep_elt_layout = sep_layout.WithRestOfLine(elt_layouts[j + 1])
        line_layout = line_layout.WithRestOfLine(sep_elt_layout)
        last_breaking = self.elements[j + 1].is_breaking
      else:  # Not executed if last_breaking
        solutions_i.append(line_layout.WithRestOfLine(rest_of_line))
      wrap_solutions[i] = support.MinSolution(solutions_i)
    # Once wrap_solutions is complete, the optimum layout for the entire block
    # is the optimum layout for the last n - 0 elements.
    return wrap_solutions[0]


class VerbBlock(LayoutBlock):
  """A block that prints out several lines of text verbatim."""

  def __init__(self, lines, is_breaking=True, first_nl=False):
    super(VerbBlock, self).__init__(is_breaking)
    self.lines = lines
    self.first_nl = first_nl

  def __repr__(self):
    return self.lines[0][:3] + '...' + self.lines[-1][-3:]

  def DoOptLayout(self, rest_of_line):
    # The solution for this block is essentially that of a TextBlock(''), with
    # an abberant layout calculated as follows.
    l_elts = []
    for i, ln in enumerate(self.lines):
      if i > 0 or self.first_nl:
        l_elts.append(support.LayoutElement.NewLine())
      l_elts.append(support.LayoutElement.String(ln))
    layout = support.Layout(l_elts)
    span = 0
    sf = support.SolutionFactory()
    if _options.m0 > 0:  # Prevent incoherent solutions
      sf.Append(0, span, 0, 0, layout)
    # _options.m1 == 0 is absurd
    sf.Append(_options.m0 - span, span, 0, _options.c0, layout)
    sf.Append(_options.m1 - span, span,
              (_options.m1 - _options.m0) * _options.c0,
              _options.c0 + _options.c1, layout)
    return sf.MkSolution()


def LineCommentBlock(comments):
  inl_words = ' '.join(re.sub(r'^\s*#+\s*', ' ', ln)
                       for ln in comments.lines).split()
  if not inl_words:
    block = IndentBlock(TextBlock('#'))
  else:
    block = IndentBlock(WrapBlock(map(TextBlock, inl_words),
                                  break_mult=_options.adj_comment, prefix='# '))
  block.is_breaking = True
  return block


def BlockCommentBlock(comments):
  stripped_lines = [re.sub(r'^\s*#', '', ln) for ln in comments.lines]
  block_lines = ['#' + ln if re.match(r'\s*#', comments.lines[i]) else ln
                 for i, ln in enumerate(stripped_lines)]
  block = VerbBlock(block_lines)
  block.is_breaking = True
  return block



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#

from dataclasses import dataclass
from typing import Optional
from typing import List

from rfmt.blocks import LineBlock as LB
from rfmt.blocks import IndentBlock as IB
from rfmt.blocks import TextBlock as TB
from rfmt.blocks import StackBlock as SB
from rfmt.blocks import WrapBlock as WB

from .utils import comments_sqlf

from .const import SQLString
from .ident import SQLIdentifier

from .node import SQLNode
from .node import SQLNodeList
from .expr import SQLExpr
from .types import SQLType
from .types import SQLNamedType


@dataclass(frozen=True)
class SQLFunction(SQLNode):
    name: SQLIdentifier
    params: SQLNodeList[SQLNode]
    retval: Optional[SQLNode]
    impl: SQLNode
    comments: List[str]

    def sqlf(self, compact):

        # Start the stack with comments
        stack = comments_sqlf(self.comments)

        # Get params as a list of sqlf
        paramf = []
        for param in self.params[:-1]:
            paramf.append(LB([param.sqlf(compact), TB(',')]))
        if self.params:
            paramf.append(self.params[-1].sqlf(compact))

        stack.append(LB([
            TB('CREATE TEMPORARY FUNCTION '),
            self.name.sqlf(True),
            TB('('),
            WB(paramf, sep=' '),
            TB(')')
        ]))

        if self.retval:
            stack.append(LB([TB('RETURNS '),
                             self.retval.sqlf(compact)]))

        if isinstance(self.impl, SQLString):
            stack.append(TB('LANGUAGE js AS'))
            stack.append(IB(LB([
                self.impl.sqlf(compact), TB(';')
            ])))
        else:
            stack.append(TB('AS'))
            stack.append(IB(LB([self.impl.sqlf(compact),
                                TB(';')])))
        stack.append(TB(''))

        return SB(stack)

    @staticmethod
    def consume(lex) -> 'Optional[SQLFunction]':
        if not (lex.consume(['CREATE', 'TEMP', 'FUNCTION']) or
                lex.consume(['CREATE', 'TEMPORARY', 'FUNCTION'])):
            return None

        comments = lex.get_comments()

        name = (SQLIdentifier.consume(lex) or
                lex.error('Expected UDF name'))

        lex.expect('(')
        params = []
        while True:
            var_name = SQLIdentifier.parse(lex)
            ltype = SQLType.parse(lex)
            params.append(SQLNamedType(var_name, ltype))
            if not lex.consume(','):
                break
        lex.expect(')')

        rtype = None

        # Javascript function
        if lex.consume('RETURNS'):
            rtype = SQLType.parse(lex)
            lex.expect('LANGUAGE')
            lex.expect('JS')
            lex.expect('AS')
            impl = (SQLString.consume(lex) or
                    lex.error('Expected Javascript code'))

        # SQL-expression
        else:
            lex.expect('AS')
            impl = SQLExpr.parse(lex)

        comments.extend(lex.get_comments())

        return SQLFunction(name, SQLNodeList(params),
                           rtype, impl, comments)



#  Copyright 2015 Google Inc. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Supporting infrastructure for the block language."""

from . import base
from builtins import str
import math

options = base.Options()  # Shorthand for convenient access


# Shorthand constant, used to denote the "virtual" knot at infinity after the
# last knot explicitly specified in a Solution object (see class definition
# below).
INFINITY = float('inf')


class Console(object):
  """An object that mediates textual output of a code layout."""

  def __init__(self, outp):
    self._m0 = options.m0
    self._m1 = options.m1
    self._h_pos = 0
    self._margins = []
    self._outp = outp

  @property
  def margin(self):
    """The offset from column 0 at which output is currently printed."""
    return self._margins[-1]

  def PrintGuide(self, initial_newline=True, m0_too=True):
    """Output debugging information delimiting the console."""
    inl = '\n' if initial_newline else ''
    print >>self._outp, inl + '=' * self._m1
    if m0_too:
      print >>self._outp, inl + '-' * self._m0

  def String(self, s):
    """Write a string on the console.

    Args:
      s: the string to be written. It is assumed that s contains no
        newline characters.
    """
    self._outp.write(str(s))
    self._h_pos += len(s)

  def Space(self, n):
    """Write a string of spaces on the console.

    Args:
      n: the number of spaces to be written (integer).
    """
    self.String(' ' * n)

  def NewLine(self, indent=True):
    """Start a new line, optionally beginning at the current margin.

    Args:
      indent: whether to preserve the current margin after the new line.
    """
    self.String('\n')
    self._h_pos = 0
    if indent:
      self.Space(self.margin)

  def NewLineSpace(self, n):
    """Start a new line, indenting from the current margin.

    Args:
      n: the number of spaces by which to indent.
    """
    self.NewLine()
    self.Space(n)

  def PrintLayout(self, layout):
    """Print a layout on the console, pushing a new margin for the duration.

    Args:
      layout: the layout object to be printed (see below for class
        definition).
    """
    self._margins.append(self._h_pos)
    layout.PrintOn(self)
    self._margins.pop()


class PrintDescriptionConsole(object):
  """A console that produces a description of the output.

  Used primarily in generating printable representations of layout objects.
  See Console class for method descriptions.
  """

  def __init__(self):
    self.out = []

  def String(self, s):
    self.out.append(s)

  def Space(self, n):
    self.String('<spc(%s)>' % n)

  def NewLine(self, indent=True):
    self.String('<NL%s>' % ('i' * indent))

  def NewLineSpace(self, n):
    self.NewLine()
    self.Space(n)

  def PrintLayout(self, layout):
    layout.PrintOn(self)

  def Output(self):
    return ''.join(self.out)


class LayoutElement(object):
  """An element of a layout object---a directive to the console.

  This class sports a collection of static methods, each of which returns
  an anonymous function invoking a method of the console to which it is
  applied.
  Refer to the corresponding methods of the Console class for descriptions of
  the methods involved.
  """

  @staticmethod
  def String(s):
    return lambda console: console.String(s)

  @staticmethod
  def NewLine(indent=True):
    return lambda console: console.NewLine(indent)

  @staticmethod
  def NewLineSpace(n):
    return lambda console: console.NewLineSpace(n)

  @staticmethod
  def PrintLayout(layout):
    return lambda console: console.PrintLayout(layout)


class Layout(object):
  """An object containing a sequence of directives to the console."""

  def __init__(self, elements):
    self.elements = elements

  def __str__(self):
    pr_cons = PrintDescriptionConsole()
    self.PrintOn(pr_cons)
    return pr_cons.Output()

  def PrintOn(self, console):
    """Have the console execute all directives in this object."""
    for e in self.elements:
      e(console)

  def __add__(self, layout):
    """Concatenate the directives in two layouts.

    Args:
      layout: the layout whose directives are to follow this one's.
    Returns:
      A new Layout, which concatenates both.
    """
    return self.__class__(self.elements + layout.elements)

  @staticmethod
  def Stack(layouts):
    """Return the vertical composition of a sequence of layouts.

    Args:
      layouts: a sequence of Layout objects.
    Returns:
      A new Layout, stacking the arguments.
    """
    l_elts = []
    for l in layouts:
      for e in l.elements:
        l_elts.append(e)
      l_elts.append(LayoutElement.NewLine())
    return Layout(l_elts[:-1])  # Drop the last NewLine()


class Solution(object):
  """An interim solution produced during layout optimization.

  A Solution object effectively maps an integer (the left margin at which the
  solution is placed) to a layout notionally optimal for that margin,
  together with cost information used to evaluate the layout. For compactness,
  the map takes the form of a piecewise-linear cost function, with associated
  layouts.

  A Solution comprises five variables:
    knots - a list of ints, specifying the margin settings at which the layout
      changes. Note that the first knot is required to be 0.
    spans - a list of ints, giving for each knot, the width of the corresponding
      layout in characters.
    intercepts - list of floats; constant cost associated with each knot.
    gradients - list of flots; at each knot, the rate with which the layout cost
      increases with an additional margin indent of 1 character.
    layouts - the Layout objects expressing the optimal layout between
      each knot.

  In addition to these items of data, a Solution object also facilitates
  iteration through the knots and the associated spans, intercepts, etc.
  """

  def __init__(self, knots, spans, intercepts, gradients, layouts):
    self.knots = list(map(int, knots))
    self.spans = list(map(int, spans))
    self.intercepts = list(map(float, intercepts))
    self.gradients = list(map(float, gradients))
    self.layouts = layouts
    self.index = 0

  def __repr__(self):
    def KnotRepr(elts):
      k, s, a, b, l = elts
      return '%d/(%d, %.2f, %.2f, %s)' % (k, s, a, b, l.__str__())
    return '<%s>' % (', '.join(map(KnotRepr,
                                   zip(self.knots, self.spans, self.intercepts,
                                       self.gradients, self.layouts))))

  # Iteration protocol
  def Reset(self):
    """Begin iteration."""
    self.index = 0

  def Advance(self):
    """Advance to the next knot."""
    self.index += 1

  def Retreat(self):
    """Move back a knot."""
    self.index -= 1

  def CurKnot(self):
    """The currently indexed knot."""
    return self.knots[self.index]

  def CurSpan(self):
    return self.spans[self.index]

  def CurIntercept(self):
    return self.intercepts[self.index]

  def CurGradient(self):
    return self.gradients[self.index]

  def CurLayout(self):
    return self.layouts[self.index]

  def CurIndex(self):
    return self.index

  def CurValueAt(self, m):
    """The value (cost) extrapolated for margin m from the current knot."""
    # Since a Solution's cost is represented by a piecewise linear function,
    # the extrapolation in this case is linear, from the current knot.
    return self.CurIntercept() + self.CurGradient() * (m - self.CurKnot())

  def NextKnot(self):
    """The knot after the once currently indexed."""
    try:
      return self.knots[self.index + 1]
    except IndexError:
      return INFINITY

  def MoveToMargin(self, m):
    """Adjust the index so m falls between the current knot and the next."""
    if self.CurKnot() > m:
      while self.CurKnot() > m:
        self.Retreat()
    else:
      while self.NextKnot() <= m:
        self.Advance()

  def PlusConst(self, const):
    """Add a constant to all values of this Solution."""
    return self.__class__(self.knots, self.spans,
                          [a + const for a in self.intercepts],
                          self.gradients, self.layouts)

  def WithRestOfLine(self, rest_of_line):
    """Return a Solution that joins the rest of the line right of this one.

    Args:
      rest_of_line: a Solution object representing the code laid out on the
        remainder of the line, or None, if the rest of the line is empty.
    Returns:
      A new Solution object juxtaposing the layout represented by this
      Solution to the immediate right of the remainder of the line.
    """
    return self if rest_of_line is None else HPlusSolution(self, rest_of_line)


class SolutionFactory(object):
  """A factory object used to construct new Solution objects.

  The factory performs basic consistency checks, and eliminates redundant
  segments that are linear extrapolations of those that precede them.
  """

  def __init__(self):
    self.entries = []

  def Append(self, knot, span, intercept, gradient, layout):
    """Add a segment to a Solution under construction."""
    if self.entries:
      # Don't add a knot if the new segment is a linear extrapolation of
      # the last.
      k_last, s_last, i_last, g_last, _ = self.entries[-1]
      if (span == s_last and gradient == g_last and
          i_last + (knot - k_last) * g_last == intercept):
        return
    if knot < 0 or span < 0 or intercept < 0 or gradient < 0:
      raise AssertionError(('Internal error: bad layout'
                            '(k %d, s %d, i %f, g %f)') % (knot, span,
                                                           intercept, gradient))
    self.entries.append((knot, span, intercept, gradient, layout))

  def MkSolution(self):
    """Construct and return a new Solution with the data in this object."""
    return Solution(*zip(*self.entries))


def HPlusSolution(s1, s2):
  """The Solution that results from joining two Solutions side-by-side.

  Args:
    s1: Solution object
    s2: Solution object
  Returns:
    A new Solution reflecting a layout in which s2 ('s layout) is placed
    immediately to the right of s1.

  The resulting Solution object maps each prospective left margin m to the
  span, cost and layout information that would result from siting Solution s1
  at m, and then placing s2 at margin m + sp1(m), where sp1(m) is the span
  of characters occupied by the layout to which s1 maps m. In general, of
  course, both s1 and s2's layouts may occupy multiple lines, in which case
  s2's layout begins at the end of the last line of s1's layout---the span
  in this case is the span of s1's last line.
  """
  col = SolutionFactory()
  s1.Reset()
  s2.Reset()
  s1_margin = 0
  s2_margin = s1.CurSpan()
  s2.MoveToMargin(s2_margin)
  while True:
    # When forming the composite cost gradient and intercept, we must
    # eliminate the over-counting of the last line of the s1, which is
    # attributable to its projection beyond the margins.
    g1 = s1.CurGradient()
    g2 = s2.CurGradient()
    overhang0 = s2_margin - options.m0  # s2_margin = m1 + span of s1
    overhang1 = s2_margin - options.m1  # s2_margin = m1 + span of s1
    g_cur = (g1 + g2 -
             options.c0 * (overhang0 >= 0) -
             options.c1 * (overhang1 >= 0))
    i_cur = (s1.CurValueAt(s1_margin) + s2.CurValueAt(s2_margin) -
             options.c0 * max(overhang0, 0) -
             options.c1 * max(overhang1, 0))
    # The Layout computed by the following implicitly sets the margin
    # for s2 at the end of the last line printed for s1.
    col.Append(s1_margin, s1.CurSpan() + s2.CurSpan(), i_cur, g_cur,
               Layout([LayoutElement.PrintLayout(s1.CurLayout()),
                       LayoutElement.PrintLayout(s2.CurLayout())]))
    # Move to the knot closest to the margin of the corresponding
    # component.
    kn1 = s1.NextKnot()
    kn2 = s2.NextKnot()
    if kn1 == INFINITY and kn2 == INFINITY: break
    # Note in the following that one of kn1 or kn2 may be infinite.
    if kn1 - s1_margin <= kn2 - s2_margin:
      s1.Advance()
      s1_margin = kn1
      s2_margin = s1_margin + s1.CurSpan()
      # Note that s1.CurSpan() may have changed, and s2_margin may
      # decrease, so we cannot simply increment s2's index.
      s2.MoveToMargin(s2_margin)
    else:
      s2.Advance()
      s2_margin = kn2
      s1_margin = s2_margin - s1.CurSpan()
  return col.MkSolution()


def VSumSolution(solutions):
  """The layout that results from stacking several Solutions vertically.

  Args:
    solutions: a non-empty sequence of Solution objects
  Returns:
    A Solution object that lays out the solutions vertically, separated by
    newlines, with the same left margin.
  """
  if len(solutions) == 1: return solutions[0]
  col = SolutionFactory()
  for s in solutions:
    s.Reset()
  margin = 0  # Margin for all components
  while True:
    col.Append(margin, solutions[-1].CurSpan(),
               sum(s.CurValueAt(margin) for s in solutions),
               sum(s.CurGradient() for s in solutions),
               Layout.Stack(s.CurLayout() for s in solutions))
    # The distance to the closest next knot from the current margin.
    d_star = min(s.NextKnot() - margin for s in solutions
                 if s.NextKnot() > margin)  # TODO(pyelland): Redundant check?
    if d_star == INFINITY:
      break
    margin += d_star
    for s in solutions:
      s.MoveToMargin(margin)
  return col.MkSolution()


def MinSolution(solutions):
  """Form the piecewise minimum of a sequence of Solutions.

  Args:
    solutions: a non-empty sequence of Solution objects
  Returns:
    values Solution object whose cost is the piecewise minimum of the Solutions
    provided, and which associates the minimum-cost layout with each piece.
  """
  if len(solutions) == 1: return solutions[0]
  factory = SolutionFactory()
  for s in solutions:
    s.Reset()
  n = len(solutions)
  k_l = 0
  last_i_min_soln = -1  # Index of the last minimum solution
  last_index = -1  # Index of the current knot in the last minimum solution
  # Move through the intervals [k_l, k_h] defined by the glb of the partitions
  # defined by each of the solutions.
  while k_l < INFINITY:
    k_h = min(s.NextKnot() for s in solutions) - 1
    gradients = [s.CurGradient() for s in solutions]
    while True:
      values = [s.CurValueAt(k_l) for s in solutions]
      # Use the index of the corresponding solution to break ties.
      min_value, min_gradient, i_min_soln = min((values[i], gradients[i], i)
                                                for i in range(n))
      min_soln = solutions[i_min_soln]
      if i_min_soln != last_i_min_soln or min_soln.CurIndex() != last_index:
        # Add another piece to the new Solution
        factory.Append(k_l, min_soln.CurSpan(), min_value, min_gradient,
                       min_soln.CurLayout())
        last_i_min_soln = i_min_soln
        last_index = min_soln.CurIndex()
      # It's possible that within the current interval, the minimum solution
      # may change, should a solution with a lower initial value but greater
      # gradient surpass the value of one with a higher initial value but
      # lesser gradient. In such instances, we need to add an extra piece to the
      # new solution.
      distances_to_cross = [math.ceil((values[i] - min_value) /
                                      (min_gradient - gradients[i]))
                            for i in range(n) if gradients[i] < min_gradient]
      # Compute positions of all crossovers in [k_l, k_h]
      crossovers = [k_l + d for d in distances_to_cross if k_l + d <= k_h]
      if crossovers:  # Proceed to crossover in [k_l, k_h]
        k_l = min(crossovers)
      else:  # Proceed to next piece
        k_l = k_h + 1
        if k_l < INFINITY:
          for s in solutions:
            s.MoveToMargin(k_l)
        break
  return factory.MkSolution()



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#


from rfmt.blocks import LineBlock as LB
from rfmt.blocks import TextBlock as TB

from .node import SQLNode


def comments_sqlf(comments):
    if not comments:
        return []
    return [TB('-- ' + c) for subc in comments for c in subc.split('\n')]


def with_commas(compact: bool, args, sep=',', tail=None):
    vargs = [a for a in args if isinstance(a, SQLNode)]

    if not vargs:
        return []

    with_commas = []
    for a in vargs[:-1]:
        with_commas.append(LB([a.sqlf(compact), TB(sep)]))
    if tail:
        with_commas.append(
            LB([vargs[-1].sqlf(compact), TB(tail)]))
    else:
        with_commas.append(vargs[-1].sqlf(compact))

    return with_commas



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#


from typing import List
from typing import Optional
from dataclasses import dataclass

from rfmt.blocks import StackBlock as SB

from .node import SQLNodeList
from .node import SQLNode

from .func import SQLFunction
from .query import SQLQuery
from .dml import SQLDML
from .cmd import SQLCommand


@dataclass(frozen=True)
class SQLScript(SQLNode):
    commands: 'SQLNodeList[SQLScript]'

    def sqlf(self, compact):
        return SB([cmd.sqlf(compact) for cmd in self.commands])

    @staticmethod
    def parse(lex) -> 'SQLScript':
        commands: List[SQLNode] = []
        while True:
            functions: List[SQLFunction] = []
            while True:
                func = SQLFunction.consume(lex)
                if not func:
                    break
                functions.append(func)

                lex.expect(';')

            lex.consume(';')

            # Consume noise
            while lex.consume('\\time') or lex.consume('\\t'):
                continue

            main_query = (SQLQuery.consume(lex) or
                          SQLDML.consume(lex) or
                          SQLCommand.consume(lex))

            if not main_query:
                break

            if functions:
                commands.append(SQLWithFunctions(main_query,
                                                 SQLNodeList(functions)))

            else:
                commands.append(main_query)

        if not commands:
            lex.error('Expect at least one SQL statement')

        lex.expect_end()

        return SQLScript(SQLNodeList(commands))


@dataclass(frozen=True)
class SQLWithFunctions(SQLQuery):
    sql: SQLNode
    funcs: Optional[SQLNode]

    def sqlf(self, compact):
        stack = []
        if self.funcs:
            for func in self.funcs:
                stack.append(func.sqlf(compact))
        stack.append(self.sql.sqlf(compact))
        return SB(stack)

    @staticmethod
    def consume(lex):
        functions: List[SQLFunction] = []

        while True:
            func = SQLFunction.consume(lex)
            if not func:
                break
            functions.append(func)

            lex.expect(';')

        lex.consume(';')

        # Consume noise
        while lex.consume('\\time') or lex.consume('\\t'):
            continue

        main_query = (SQLQuery.consume(lex) or
                      SQLDML.consume(lex) or
                      SQLCommand.consume(lex))

        if not main_query:
            return None

        if functions:
            return SQLWithFunctions(main_query, SQLNodeList(functions))

        return main_query



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#

from dataclasses import dataclass
from typing import Optional
from typing import Tuple
from typing import List

from rfmt.blocks import LineBlock as LB
from rfmt.blocks import TextBlock as TB
from rfmt.blocks import StackBlock as SB
from rfmt.blocks import IndentBlock as IB
from rfmt.blocks import ChoiceBlock as CB
from rfmt.blocks import WrapBlock as WB

from .utils import with_commas

from .types import SQLType
from .query import SQLQuery
from .query import SQLNamedTable
from .query_impl import SQLFrom
from .expr import SQLExpr
from .ident import SQLIdentifierPath
from .node import SQLNode
from .node import SQLNodeList
from .query import SQLTableSource


@dataclass(frozen=True)
class SQLDML(SQLNode):

    @staticmethod
    def consume(lex):
        return (SQLInsert.consume(lex) or
                SQLUpdate.consume(lex) or
                SQLDelete.consume(lex) or
                SQLCreate.consume(lex))


@dataclass(frozen=True)
class SQLInsert(SQLDML):
    table: SQLTableSource
    fields: Optional[SQLNodeList[SQLNode]]
    sql: SQLNode

    def sqlf(self, compact):
        # TODO(scannell) Add compact version
        if not self.fields:
            return SB([
                LB([
                    TB('INSERT INTO '),
                    self.table.sqlf(compact=True),
                    TB(' ('),
                ]),
                IB(self.sql.sqlf(compact)),
                TB(');'),
            ])
        return SB([
            LB([
                TB('INSERT INTO '),
                self.table.sqlf(compact=True),
                TB(' ('),
            ]),
            IB(
                SB(with_commas(compact, self.fields, tail=')'))
            ),
            self.sql.sqlf(compact),
            TB(';'),
        ])

    @staticmethod
    def consume(lex) -> 'Optional[SQLInsert]':
        if not lex.consume('INSERT'):
            return None

        lex.consume('INTO')

        table = SQLNamedTable.parse(lex, is_write=True)

        lex.expect('(')

        # It's possible to have a SQL statement in here.
        # .. this is with an implicit column list.
        query = SQLQuery.consume(lex)
        if query:
            lex.expect(')')
            return SQLInsert(table, None, query)

        # Pull out the values
        cols = []
        while True:
            cols.append(SQLIdentifierPath.parse(lex))
            if not lex.consume(','):
                break
        lex.expect(')')

        # Now the SQL
        query = SQLQuery.consume(lex)
        if query:
            return SQLInsert(table, SQLNodeList(cols), query)

        lex.error('Insert with explicit values unsupported')
        return None


@dataclass(frozen=True)
class SQLDelete(SQLDML):
    table: SQLNode
    where_expr: SQLNode

    def sqlf(self, compact):
        delete = [TB('DELETE FROM '), self.table.sqlf(compact)]
        return SB([
            LB(delete),
            LB([
                TB('WHERE '),
                self.where_expr.sqlf(compact),
            ]),
            TB(';'),
        ])

    @staticmethod
    def consume(lex) -> 'Optional[SQLDelete]':
        if not lex.consume('DELETE'):
            return None

        lex.consume('FROM')

        table = SQLNamedTable.parse(lex, is_write=True)

        lex.expect('WHERE')

        where_expr = SQLExpr.parse(lex)

        return SQLDelete(table, where_expr)


@dataclass(frozen=True)
class SQLUpdate(SQLDML):
    table_name: SQLNode
    update_fields: SQLNodeList[SQLIdentifierPath]
    update_exprs: SQLNodeList[SQLNode]
    from_tables: SQLNode
    where_expr: SQLNode

    # TODO(scannell) honour compact mode
    def sqlf(self, compact):
        tname = [TB('UPDATE '), self.table_name.sqlf(compact)]

        updatestmt = [
            LB(tname),
            TB('SET'),
        ]
        for i in range(len(self.update_fields)):
            field = self.update_fields[i]
            expr = self.update_exprs[i]
            if i == len(self.update_fields)-1:
                updatestmt.append(LB([
                    field.sqlf(False), TB(' = '), expr.sqlf(False)
                ]))
            else:
                updatestmt.append(LB([
                    field.sqlf(False), TB(' = '), expr.sqlf(False),
                    TB(',')
                ]))
        if self.from_tables:
            updatestmt.append(LB([
                TB(' FROM '),
                self.from_tables.sqlf(False)
            ]))
        if self.where_expr:
            updatestmt.append(LB([
                TB(' WHERE '),
                self.where_expr.sqlf(False)
            ]))
        return SB(updatestmt)

    @staticmethod
    def consume(lex) -> 'Optional[SQLUpdate]':
        if not lex.consume('UPDATE'):
            return None

        table_name = SQLNamedTable.parse(lex, is_write=True)

        lex.expect('SET')

        update_fields = []
        update_exprs = []
        while True:
            update_fields.append(SQLIdentifierPath.parse(lex))
            lex.expect('=')
            update_exprs.append(SQLExpr.parse(lex))
            if not lex.consume(','):
                break

        from_tables = lex.consume('FROM') and SQLFrom.parse(lex)
        where_expr = lex.consume('WHERE') and SQLExpr.parse(lex)

        return SQLUpdate(table_name,
                         SQLNodeList(update_fields),
                         SQLNodeList(update_exprs),
                         from_tables, where_expr)


@dataclass(frozen=True)
class SQLOption(SQLDML):
    key: str
    val: SQLExpr

    def sqlf(self, compact):
        compact_sql = LB([TB(key), TB('='), TB(self.val.sqlf(True))])
        if compact:
            return compact_sql

        return CB([
            compact_sql,
            SB([
                LB([TB(key), TB(' =')]),
                IB(self.val.sqlf(False)),
            ])
        ])

    @staticmethod
    def parse(lex) -> 'SQLOption':
        key = lex.consumer_identifier()
        lex.expect('=')
        val = SQLExpr.parse(lex)

        return SQLOption(key, val)


@dataclass(frozen=True)
class SQLOptions(SQLDML):
    options: SQLNodeList[SQLOption]

    def sqlf(self):
        lines = [TB('OPTIONS(')]
        lines.extend(with_commas(True, self.options))
        lines.append(TB(')'))

        if compact:
            return LB(lines)

        stack = [TB('OPTIONS(')]
        stack.extend(IB(WB(with_commas(False, self.options))))
        stack.append(TB(')'))

        return CB([
            LB(lines),
            SB(stack),
        ])

    @staticmethod
    def consume(lex) -> 'Optional[SQLOptions]':
        if not lex.consume('OPTIONS'):
            return None

        lex.expect('(')
        options: List[SQLOption] = []
        while True:
            options.append(SQLOption.parse(lex))
            if not lex.consume(','):
                break
        lex.expect(')')

        return SQLOptions(SQLNodeList(options))


@dataclass(frozen=True)
class SQLColumn(SQLDML):
    name: str
    col_type: SQLType
    nullable: bool
    opts: Optional[SQLOptions]

    def sqlf(self, compact):
        lines = [TB(self.name), TB(' '),
                 self.col_type.sqlf(True)]
        if not self.nullable:
            lines.append(TB(' NOT NULL'))
        if self.opts:
            lines.append(TB(' '))
            lines.append(self.opts.sqlf(True))

        if compact:
            return LB(lines)

        # Indented block for non-compact
        iblock = [self.col_type.sqlf(False)]
        if not self.nullable:
            iblock.append(TB('NOT NULL'))
        if self.opts:
            iblock.append(self.opts.sqlf(False))

        return CB([
            LB(lines),
            SB([
                TB(self.name),
                IB(SB(iblock)),
            ])
        ])

    @staticmethod
    def parse(lex) -> 'SQLColumn':
        name = (lex.consume_identifier() or
                lex.error('expected column name'))
        col_type = SQLType.parse(lex)

        # Get whether nullable
        nullable = True
        if lex.consume('NOT'):
            lex.expect('NULL')
            nullable = False

        # Any options
        opts = SQLOptions.consume(lex)

        return SQLColumn(
            name, col_type,
            nullable, opts
        )


@dataclass(frozen=True)
class SQLCreate(SQLDML):
    clause: str
    table: SQLTableSource
    columns: SQLNodeList[SQLColumn]
    options: Optional[SQLOptions]
    query: Optional[SQLNode]

    def sqlf(self, compact):
        lines = [TB(self.clause), TB(' ')]
        lines.append(self.table.sqlf(True))
        if self.columns:
            lines.append(TB(' ('))
            lines.extend(with_commas(True, self.columns))
            lines.append(TB(')'))
        if self.options:
            lines.append(TB(' '))
            lines.append(self.options.sqlf(True))
        if self.query:
            lines.append(TB(' AS '))
            lines.append(self.query.sqlf(True))

        if compact:
            return LB(lines)

        stack = [LB([TB(self.clause),TB(' '),self.table.sqlf(True)])]
        if self.columns:
            stack.extend([
                TB('('),
                IB(WB(
                    with_commas(False, self.columns),
                )),
                TB(')'),
            ])
        if self.options:
            stack.append(self.options.sqlf(False))
        if self.query:
            stack.extend([
                TB('AS'),
                self.query.sqlf(False),
            ])

        return SB(stack)

    @staticmethod
    def consume(lex):
        if not lex.consume('CREATE'):
            return None

        #  Find CFREATE OR REPLACE TABLE clause
        clause = None
        if lex.consume(['TEMP', 'TABLE']):
            clause = 'CREATE TEMP TABLE'
        elif lex.consume('OR'):
            lex.expect('REPLACE')
            lex.expect('TABLE')
            clause = 'CREATE OR REPLACE TABLE'
        else:
            lex.expect('TABLE')

            # Check for IF NOT EXISTS
            if lex.consume('IF'):
                lex.expect('NOT')
                lex.expect('EXISTS')
                clause = 'CREATE TABLE IF NOT EXISTS'
            else:
                clause = 'CREATE TABLE'

        # Parse table
        table = SQLNamedTable.parse_no_alias(lex, is_write=True)

        # Parse columns
        columns: List[SQLColumn] = []
        if lex.consume('('):
            while True:
                columns.append(SQLColumn.parse(lex))
                if not lex.consume(','):
                    break
            lex.expect(')')

        partition_by = None
        if lex.consume('PARTITION'):
            lex.expect('BY')
            # TODO: Finish
            lex.error('PARTITION BY unimplemented')

        cluster_by = None
        if lex.consume('CLUSTER'):
            lex.expect('BY')
            # TODO: Finish
            lex.error('CLUSTER BY unimplemented')

        # Find options
        options = SQLOptions.consume(lex)

        # Find query-for-creation (if any)
        query = None
        if lex.consume('AS'):
            query = SQLQuery.parse(lex)

        return SQLCreate(clause, table, columns, options, query)





# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#


import re

from dataclasses import dataclass
from dataclasses import fields
from dataclasses import replace

from typing import Generic
from typing import TypeVar
from typing import Sequence
from typing import Any

from io import StringIO

import rfmt.base as base


# Default options for formatting SQL
_options: Any = base.Options()
_options.m0 = 0
_options.m1 = 100
_options.c0 = 0.05
_options.c1 = 240
_options.cb = 2
_options.cpack = 1e-3
_options.ind = 2
_options.force_brace = False
_options.space_arg_eq = True


@dataclass(frozen=True)
class SQLNode:

    def __str__(self):
        """Minified SQL for this node."""
        sqlstr = self.as_sql(True)
        # Remove comments
        sqlstr = re.sub(r'^\s*--[^\n]*\n', '', sqlstr)
        # Remove multispaces to a single space
        sqlstr = re.sub(r'\s+', ' ', sqlstr)
        # Needed twice for single-letter operators
        sqlstr = re.sub(r'([^\w\']) ([^-])', '\\1\\2', sqlstr)
        sqlstr = re.sub(r'([^\w\']) ([^-])', '\\1\\2', sqlstr)
        # Needed twice for single-letter operators
        sqlstr = re.sub(r'([^-]) ([^\w\'])', '\\1\\2', sqlstr)
        sqlstr = re.sub(r'([^-]) ([^\w\'])', '\\1\\2', sqlstr)
        # Remove space at end
        sqlstr = re.sub(r'\s*$', '', sqlstr)
        return sqlstr

    def get_tree(self):
        """Return compact representation of tree"""
        return repr(self)

    def as_sql(self, compact=False):
        """Fully formatted SQL for this node."""
        outp = StringIO()
        self.sqlf(compact).PrintOn(outp)
        return re.sub(r' *$', '', outp.getvalue(), flags=re.MULTILINE)

    def sqlf(self, compact):
        """Return format-blocks for this node."""
        del compact  # Unused
        raise Exception('Abstract method not implemented in {}.'.format(
                        self.get_name()))

    def get_name(self):
        return self.__class__.__name__

    def children(self):
        for field in fields(self):
            val = getattr(self, field.name)
            if val is None:
                continue
            if issubclass(type(val), SQLNode):
                yield (field.name, val)
            continue

    def rewrite_tree(self, fopt):
        nself = fopt(self) or self

        nvals = dict()
        for name, val in nself.children():
            nvals[name] = val.rewrite_tree(fopt)

        return replace(nself, **nvals)


SQLNodeType = TypeVar('SQLNodeType', bound=SQLNode)


class SQLNodeList(SQLNode, Generic[SQLNodeType], tuple):
    def __new__(cls, val: Sequence[SQLNodeType]):
        return tuple.__new__(cls, val)

    def __init__(self, val):
        del val  # Unused  (see __new__)
        super().__init__()

    def children(self):
        for i, child in enumerate(self):
            yield (i, child)

    def __repr__(self):
        return tuple.__repr__(self)

    def rewrite_tree(self, fopt):
        return SQLNodeList([f.rewrite_tree(fopt) for f in self])
